import dataclasses
import enum
import math
import unittest
import uuid
from io import BytesIO
from typing import Optional

import numpy as np

from hippolyzer.lib.base.datatypes import *
import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.llanim import Animation, Joint, RotKeyframe
from hippolyzer.lib.base.namevalue import NameValuesSerializer, NameValueSerializer


class BaseSerializationTest(unittest.TestCase):
    def setUp(self) -> None:
        self.writer = se.BufferWriter("!")

    def _get_reader(self, pod=False):
        return se.BufferReader(self.writer.endianness, self.writer.copy_buffer(), pod)

    def _assert_coords_fuzzy_equals(self, expected, val):
        if len(expected) != len(expected):
            # Take advantage of existing formatting, this will fail
            self.assertSequenceEqual(expected, val)
        for x, y in zip(expected, val):
            if not math.isclose(x, y, rel_tol=1e-4):
                self.assertSequenceEqual(expected, val)


class SerializationTests(BaseSerializationTest):
    def test_basic(self):
        self.writer.write(se.U32, 1)
        self.writer.write(se.U16, 2)
        reader = self._get_reader()
        self.assertEqual(reader.read(se.U32), 1)
        self.assertEqual(reader.read(se.U16), 2)
        # No more data left
        self.assertFalse(reader)

    def test_peeking(self):
        self.writer.write(se.U32, 1)
        reader = self._get_reader()
        self.assertEqual(reader.read(se.U32, peek=True), 1)
        self.assertEqual(reader.read(se.U32, peek=True), 1)
        # Should still be data left
        self.assertTrue(reader)

    def test_basic_collection(self):
        orig_list = [1, 2, 3, 4]
        coll = se.Collection(se.U8, se.U16)
        self.writer.write(coll, orig_list)
        reader = self._get_reader()
        self.assertListEqual(reader.read(coll), orig_list)
        self.assertFalse(reader)

    def test_greedy_collection(self):
        orig_list = [1, 2, 3, 4]
        coll = se.Collection(None, se.U16)
        self.writer.write(coll, orig_list)
        reader = self._get_reader()
        self.assertListEqual(reader.read(coll), orig_list)
        self.assertFalse(reader)

    def test_overly_large_collection_raises(self):
        orig_list = [1] * 300
        coll = se.Collection(se.U8, se.U16)
        with self.assertRaises(ValueError):
            self.writer.write(coll, orig_list)

    def test_strings(self):
        self.writer.write(se.Str(se.U8), "foobar")
        self.writer.write(se.Str(se.U16), "baz")
        reader = self._get_reader()
        self.assertEqual(reader.read(se.Str(se.U8)), "foobar")
        self.assertEqual(reader.read(se.Str(se.U16)), "baz")

    def test_prim_ranges(self):
        self.assertEqual(se.S32.max_val, 2147483647)
        self.assertEqual(se.S32.min_val, -2147483648)
        self.assertEqual(se.U32.max_val, 4294967295)
        self.assertEqual(se.U32.min_val, 0)
        self.assertEqual(se.S8.max_val, 127)
        self.assertEqual(se.S8.min_val, -128)

    def test_fixed_strings(self):
        spec = se.StrFixed(5)
        self.writer.write(spec, "foo")
        self.writer.write(spec, "quuxy")
        with self.assertRaises(ValueError):
            self.writer.write(spec, "verylong")
        reader = self._get_reader()
        self.assertEqual(reader.read(spec), "foo")
        self.assertEqual(reader.read(spec), "quuxy")

    def test_fixed(self):
        self.writer.write(se.BytesFixed(3), b"foo")
        with self.assertRaises(ValueError):
            self.writer.write(se.BytesFixed(3), b"foobar")
        reader = self._get_reader()
        self.assertEqual(reader.read(se.BytesFixed(3)), b"foo")

    def test_uuid(self):
        val = uuid.uuid4()
        self.writer.write(se.UUID, val)
        reader = self._get_reader()
        self.assertEqual(reader.read(se.UUID), val)

    def test_uuid_pod(self):
        val = uuid.uuid4()
        self.writer.write(se.UUID, val)
        reader = self._get_reader(pod=True)
        self.assertEqual(reader.read(se.UUID), str(val))

    def test_template(self):
        template = se.Template({
            "num_1": se.U64,
            "some_nums": se.Collection(se.U8, se.U16),
            "test_str": se.Str(se.U8),
        })

        vals = {
            "num_1": 2,
            "some_nums": [1, 4, 3],
            "test_str": "hi hello",
        }
        self.writer.write(template, vals)
        reader = self._get_reader()
        self.assertDictEqual(reader.read(template), vals)

    def test_cstr(self):
        self.writer.write(se.CStr(), "foobaz")
        self.writer.write(se.U8, 1)
        reader = self._get_reader()
        self.assertEqual(reader.read(se.CStr()), "foobaz")
        self.assertEqual(reader.read(se.U8), 1)

    def test_template_size(self):
        templ = se.Template({
            "foo": se.UUID,
            "bar": se.S32,
        })
        self.assertEqual(templ.calc_size(), 20)

    def test_enum_switch(self):
        class Foo(enum.IntEnum):
            STR = 0
            U16 = 1

        template = se.EnumSwitch(se.IntEnum(Foo, se.U8), {
            Foo.STR: se.CStr(),
            Foo.U16: se.U16,
        })

        self.writer.write(template, (Foo.STR, "foo"))
        self.writer.write(template, (Foo.U16, 12))
        reader = self._get_reader()
        self.assertSequenceEqual(reader.read(template), (Foo.STR, "foo"))
        self.assertSequenceEqual(reader.read(template), (Foo.U16, 12))

    def test_enum_switch_pod(self):
        class Foo(enum.IntEnum):
            STR = 0
            U16 = 1

        template = se.EnumSwitch(se.IntEnum(Foo, se.U8), {
            Foo.STR: se.CStr(),
            Foo.U16: se.U16,
        })

        self.writer.write(template, ("STR", "foo"))
        reader = self._get_reader(pod=True)
        self.assertSequenceEqual(reader.read(template), ("STR", "foo"))

    def test_flag_switch(self):
        class Foo(enum.IntFlag):
            STR = 1
            U16 = enum.auto()
            U32 = enum.auto()
            U64 = enum.auto()

        template = se.FlagSwitch(se.IntFlag(Foo, se.U8), {
            Foo.STR: se.CStr(),
            Foo.U16: se.U16,
            Foo.U32: se.U32,
            # U64 intentionally missing
        })

        expected = {
            Foo.STR: "barbaz",
            Foo.U32: 20000,
        }

        self.writer.write(template, expected)
        reader = self._get_reader()
        self.assertSequenceEqual(reader.read(template), expected)

    def test_flag_switch_pod(self):
        class Foo(enum.IntFlag):
            STR = 1
            U16 = enum.auto()
            U32 = enum.auto()
            U64 = enum.auto()

        template = se.FlagSwitch(se.IntFlag(Foo, se.U8), {
            Foo.STR: se.CStr(),
            Foo.U16: se.U16,
            Foo.U32: se.U32,
            # U64 intentionally missing
        })

        expected = {
            "STR": "barbaz",
            "U32": 20000,
        }

        self.writer.write(template, expected)
        reader = self._get_reader(pod=True)
        self.assertSequenceEqual(reader.read(template), expected)

    def test_length_switch(self):
        template = se.LengthSwitch({
            4: se.Collection(2, se.U16),
            2: se.U16,
        })

        self.writer.write(se.U16, 1)
        reader = self._get_reader()
        self.assertEqual(reader.read(template), TaggedUnion(2, 1))
        self.writer.write(se.U8, 0)
        reader = self._get_reader()
        with self.assertRaises(KeyError):
            reader.read(template)
        self.writer.write(se.U8, 2)
        reader = self._get_reader()
        self.assertSequenceEqual(reader.read(template), TaggedUnion(4, [1, 2]))

    def test_length_switch_catch_all(self):
        template = se.LengthSwitch({
            2: se.U16,
            None: se.Null
        })

        self.writer.write(se.U32, 1)
        read_template = self._get_reader().read(template)
        # Real length returned, but catchall template should be used for both read and write.
        self.assertEqual(read_template, TaggedUnion(4, None))
        self.writer.buffer.clear()
        self.writer.write(template, read_template)
        self.assertEqual(len(self.writer), 0)

    def test_tuple_coords(self):
        self.writer.write(se.Vector3D, (0.0, 1.0, 2.0))
        reader = self._get_reader()
        self.assertEqual(len(reader), 8 * 3)
        self._assert_coords_fuzzy_equals(reader.read(se.Vector3D()), (0.0, 1.0, 2.0))
        self.assertFalse(reader)

    def test_vector3_quat(self):
        expected = Quaternion(0.08283, 0.19996, -0.37361, 0.90198)
        self.writer.write(se.PackedQuat(se.Vector3), expected)
        reader = self._get_reader()
        self.assertEqual(len(reader), 4 * 3)
        self._assert_coords_fuzzy_equals(reader.read(se.PackedQuat(se.Vector3)), expected)
        self.assertFalse(reader)

    def test_vector3_quat_pod(self):
        expected = (0.08283, 0.19996, -0.37361, 0.90198)
        self.writer.write(se.PackedQuat(se.Vector3), expected)
        reader = self._get_reader(pod=True)
        self.assertEqual(len(reader), 4 * 3)
        self._assert_coords_fuzzy_equals(reader.read(se.PackedQuat(se.Vector3)), expected)
        self.assertFalse(reader)

    def test_int_enum(self):
        class Foo(enum.IntEnum):
            bar = 1
            baz = 2

        packed = se.IntEnum(Foo, se.U8, strict=False)
        self.writer.write(packed, "bar")
        # Allowed to write invalid enum vals
        self.writer.write(packed, 3)
        with self.assertRaises(KeyError):
            # Unknown lookup strs are bad though
            self.writer.write(packed, "quuxy")

        reader = self._get_reader()
        self.assertEqual(reader.read(packed), Foo.bar)
        # not strict, will return int for unrecognized
        self.assertEqual(reader.read(packed), 3)
        self.assertFalse(reader)

    def test_int_flag(self):
        class Foo(enum.IntFlag):
            bar = 1
            baz = enum.auto()
            quux = enum.auto()

        packed = se.IntFlag(Foo, se.U8)
        self.writer.write(packed, ("bar", "quux"))
        # Allowed to write invalid enum vals
        self.writer.write(packed, 3)
        with self.assertRaises(KeyError):
            # Unknown lookup strs are bad though
            self.writer.write(packed, ("quuxy",))

        reader = self._get_reader()
        self.assertEqual(reader.read(packed), Foo.bar | Foo.quux)
        self.assertEqual(reader.read(packed), Foo.bar | Foo.baz)
        self.assertFalse(reader)

    def test_int_flag_pod(self):
        class Foo(enum.IntFlag):
            bar = 1
            baz = 2
            quux = 4

        packed = se.IntFlag(Foo, se.U8)
        self.writer.write(packed, Foo.bar | Foo.quux | 8)
        reader = self._get_reader(pod=True)
        self.assertEqual(reader.read(packed), ("bar", "quux", 8))

    def test_bit_field(self):
        bitfield = se.BitField(se.U32, {"bar": 31, "foo": 1})
        expected = {"foo": 1, "bar": 2}
        self.writer.write(bitfield, expected)
        self.writer.write(se.U32, 2147483649)

        reader = self._get_reader()
        self.assertEqual(reader.read(bitfield), expected)
        self.assertEqual(reader.read(bitfield), {"foo": 1, "bar": 1})

    def test_bitfield_dataclass(self):
        class SomeEnum(enum.IntEnum):
            FOO = 0
            BAR = 1
            BAZ = 2

        @dataclasses.dataclass
        class SomeDataclass:
            some: SomeEnum = se.bitfield_field(bits=2, adapter=se.IntEnum(SomeEnum))
            number: int = se.bitfield_field(bits=30)

        bitfield = se.BitfieldDataclass(SomeDataclass, se.U32)

        expected = SomeDataclass(some=SomeEnum.BAR, number=200)
        self.writer.write(bitfield, expected)

        reader = self._get_reader()
        self.assertEqual(reader.read(bitfield), expected)

        # Now do the POD case
        self.writer.clear()
        self.writer.write(bitfield, expected)

        reader = self._get_reader(pod=True)
        self.assertEqual(reader.read(bitfield), {
            "some": "BAR",
            "number": 200,
        })

    def test_optional_prefixed(self):
        template = se.OptionalPrefixed(se.U8)
        self.writer.write(template, None)
        self.writer.write(template, 20)

        reader = self._get_reader()
        self.assertEqual(reader.read(template), None)
        self.assertEqual(reader.read(template), 20)

    def test_optional_flagged(self):
        class Foo(enum.IntFlag):
            STR = 1
            U16 = enum.auto()
            U32 = enum.auto()
            U64 = enum.auto()

        flag_spec = se.IntFlag(Foo, se.U8)
        template = se.Template({
            "Flags": flag_spec,
            "String1": se.OptionalFlagged("Flags", flag_spec, Foo.STR, se.CStr()),
            "Int1": se.OptionalFlagged("Flags", flag_spec, Foo.U16, se.U16),
            "Int2": se.OptionalFlagged("Flags", flag_spec, Foo.U32, se.U32),
            # U64 intentionally missing
        })

        val = {
            "Flags": Foo.STR | Foo.U16 | Foo.U64,
            "String1": "barbaz",
            "Int1": 20000,
        }

        self.writer.write(template, val)
        reader = self._get_reader()
        self.assertSequenceEqual(reader.read(template), {"Int2": None, **val})

    def test_optional_flagged_pod(self):
        class Foo(enum.IntFlag):
            STR = 1
            U16 = enum.auto()
            U32 = enum.auto()
            U64 = enum.auto()

        flag_spec = se.IntFlag(Foo, se.U8)
        template = se.Template({
            "Flags": flag_spec,
            "String1": se.OptionalFlagged("Flags", flag_spec, Foo.STR, se.CStr()),
            "Int1": se.OptionalFlagged("Flags", flag_spec, Foo.U16, se.U16),
            "Int2": se.OptionalFlagged("Flags", flag_spec, Foo.U32, se.U32),
            # U64 intentionally missing
        })

        val = {
            "Flags": ("STR", "U16", "U64"),
            "String1": "barbaz",
            "Int1": 20000,
        }

        self.writer.write(template, val)
        reader = self._get_reader(pod=True)
        self.assertSequenceEqual(reader.read(template), {"Int2": None, **val})

    def test_typed_bytearray(self):
        template = se.Template({
            "Int1": se.U32,
        })
        arr_template = se.TypedByteArray(se.U32, template)
        self.writer.write(arr_template, {"Int1": 1})
        # len field + int1
        self.assertEqual(len(self.writer), 8)
        self.assertEqual(self._get_reader().read(arr_template), {"Int1": 1})

        # trailing bytes left unparsed inside the array should fail
        self.writer.buffer.clear()
        self.writer.write(se.U32, 5)
        self.writer.write_bytes(b"x" * 5)
        with self.assertRaises(ValueError):
            print(self._get_reader().read(arr_template))

    def test_parse_context(self):
        test_self = self

        class Foo(se.SerializableBase):
            def __init__(self):
                self.ser_called = False
                self.deser_called = False

            def serialize(self, val, writer: se.BufferWriter, ctx: Optional[se.ParseContext]):
                test_self.assertEqual(ctx["bar"], 1)
                test_self.assertEqual(ctx._["plugh"], 2)
                self.ser_called = True

            def deserialize(self, reader: se.BufferReader, ctx: Optional[se.ParseContext]):
                test_self.assertEqual(ctx["bar"], 1)
                test_self.assertEqual(ctx._["plugh"], 2)
                self.deser_called = True

        foo_spec = Foo()
        template = se.Template({
            "plugh": se.U16,
            "quux": se.Template({
                "bar": se.U16,
                "baz": foo_spec,
            })
        })
        self.writer.write(template, {"plugh": 2, "quux": {"bar": 1, "baz": None}})
        self._get_reader().read(template)
        self.assertTrue(foo_spec.deser_called)
        self.assertTrue(foo_spec.ser_called)

    def test_multidict(self):
        test_list = [
            (1, 2),
            (1, 3),
            (4, 3),
            (1, 4),
        ]
        spec = se.Collection(se.U32, se.Tuple(se.U8, se.U32))
        multi_spec = se.MultiDictAdapter(spec)
        self.writer.write(spec, test_list)
        written_buff = self.writer.copy_buffer()
        reader = self._get_reader()
        deser = reader.read(multi_spec)
        item_view = deser.items(multi=True)
        for i in range(2):
            self.assertEqual(list(item_view), test_list)

        self.writer.clear()
        self.writer.write(multi_spec, deser)
        self.assertEqual(written_buff, self.writer.copy_buffer())

    def test_dataclass(self):
        @dataclasses.dataclass
        class Foobar:
            foo: int = se.dataclass_field(se.U8)
            bar: str = se.dataclass_field(se.CStr())

        spec = se.Dataclass(Foobar)
        inst = Foobar(foo=1, bar="log off")

        self.writer.write(spec, inst)
        reader = self._get_reader()
        deser: Foobar = reader.read(spec)
        self.assertEqual(inst, deser)
        self.assertEqual(deser.bar, "log off")


class FileSerializationTests(BaseSerializationTest):
    def _get_reader(self, pod=False):
        return se.FHReader(self.writer.endianness, BytesIO(self.writer.copy_buffer()), pod=pod)

    def test_simple(self):
        self.writer.write(se.CStr(), "foobar")
        self.writer.write(se.CStr(), "baz")
        reader = self._get_reader()
        self.assertEqual(reader.read(se.CStr()), "foobar")
        self.assertEqual(reader.read(se.CStr(), peek=True), "baz")
        self.assertTrue(reader)
        self.assertEqual(len(reader), 4)
        self.assertEqual(reader.read(se.CStr()), "baz")
        self.assertFalse(reader)


class QuantizedFloatSerializationTests(BaseSerializationTest):
    def _test_quantized_float_roundtrips(self, prim: se.SerializablePrimitive):
        test_ranges = [
            se.QuantizedFloat(prim, 0.0, 1.0),
            # Make sure we test the 0.0 median point special-casing
            se.QuantizedFloat(prim, -1.0, 1.0),
            # Lopsided, with zero in the middle-ish
            se.QuantizedFloat(prim, -2.0, 1.0),
        ]
        for packed in test_ranges:
            for i in range(prim.min_val, prim.max_val):
                self.writer.buffer = bytearray()
                self.writer.write(prim, i)
                initial_buf = self.writer.copy_buffer()
                reader = self._get_reader()
                first_read = reader.read(packed)
                self.writer.buffer.clear()
                self.writer.write(packed, first_read)
                self.assertEqual(initial_buf, self.writer.copy_buffer())

    def test_quantized_u8_roundtrips(self):
        self._test_quantized_float_roundtrips(se.U8)

    def test_quantized_s8_roundtrips(self):
        self._test_quantized_float_roundtrips(se.S8)

    @unittest.skip("expensive")
    def test_quantized_u16_roundtrips(self):
        self._test_quantized_float_roundtrips(se.U16)

    @unittest.skip("expensive")
    def test_quantized_s16_roundtrips(self):
        self._test_quantized_float_roundtrips(se.S16)

    def test_quantized_tuplecoords(self):
        packed = se.Vector3U16(-2.0, 2.0)
        self.writer.write(packed, (-2.0, 1.0, 2.0))
        self.writer.write(packed, (-2.0, 1.0, 1.0))
        reader = self._get_reader()
        self.assertEqual(len(reader), 2 * 6)
        self._assert_coords_fuzzy_equals(reader.read(packed), (-2.0, 1.0, 2.0))
        with self.assertRaises(AssertionError):
            self._assert_coords_fuzzy_equals(reader.read(packed), (-2.0, 1.0, 2.0))
        self.assertFalse(reader)

    def test_quantized_tuple_comes_through_zero(self):
        test_val = b"\x7f\xff\x7f\xff\x7f\xff"
        packed = se.Vector3U16(-64.0, 64.0)
        self.writer.write_bytes(test_val)
        reader = self._get_reader()
        first_read = reader.read(packed)
        self.assertEqual(first_read, Vector3(0, 0, 0))
        self.writer.buffer.clear()
        self.writer.write(packed, first_read)
        self.assertEqual(self.writer.copy_buffer(), test_val)

    def test_quantized_tuplecoords_component_scales(self):
        packed = se.Vector3U16(component_scales=(
            (0.0, 1.0),
            (0.0, 1.0),
            (0.0, 4096.0)
        ))
        self.writer.write(packed, (0.0, 1.0, 4095.0))
        reader = self._get_reader()
        self.assertEqual(len(reader), 6)
        self._assert_coords_fuzzy_equals(reader.read(packed), (0.0, 1.0, 4095.0))
        self.assertFalse(reader)

    def test_quantized_tuple_signed(self):
        class Vector3S16(se.QuantizedTupleCoord):
            ELEM_SPEC = se.S16
            NUM_ELEMS = 3
            COORD_CLS = Vector3

        vec_ser = Vector3S16(-2.0, 2.0)
        init_buf = b"\x80\x00\x00\x00\x7f\xff"
        self.writer.write_bytes(init_buf)
        reader = self._get_reader()
        self.assertEqual(len(reader), 6)
        unpacked = reader.read(vec_ser)
        self._assert_coords_fuzzy_equals(unpacked, (-2.0, 0.0, 2.0))
        self.assertFalse(reader)
        self.writer.buffer.clear()
        self.writer.write(vec_ser, unpacked)
        self.assertEqual(self.writer.copy_buffer(), init_buf)

    def test_quantized_extremes(self):
        spec = se.QuantizedFloat(se.S16, -2.0, 1.0)
        self.writer.write_bytes(b"\x80\x00\x7f\xff")
        reader = self._get_reader()
        self.assertEqual(-2.0, reader.read(spec))
        self.assertEqual(1.0, reader.read(spec))

    def test_fixed_point_tuplecoord(self):
        expected_bytes = b"\xff\x80\x00\x00\x7f\x7f"
        spec = se.FixedPointVector3U16(8, 7, signed=True)
        self.writer.write_bytes(expected_bytes)
        vec: Vector3 = self._get_reader().read(spec)
        self._assert_coords_fuzzy_equals(tuple(vec), (255.0, -256.0, -1.0078))
        self.writer.clear()
        self.writer.write(spec, vec)
        self.assertEqual(expected_bytes, self.writer.copy_buffer())


class NameValueSerializationTests(BaseSerializationTest):
    EXAMPLE_NAMEVALUES = 'DisplayName STRING RW DS 𝔲𝔫𝔦𝔠𝔬𝔡𝔢𝔫𝔞𝔪𝔢\n' \
                         'FirstName STRING RW DS firstname\n' \
                         'LastName STRING RW DS Resident\n' \
                         'Title STRING RW DS foo'.encode("utf8")

    def test_basic(self):
        val = self.EXAMPLE_NAMEVALUES
        self.writer.write_bytes(val)
        reader = self._get_reader(pod=False)
        deser = reader.read(NameValuesSerializer)

        self.assertEqual(deser.to_dict()['Title'], 'foo')
        self.assertEqual(deser.to_dict()['DisplayName'], '𝔲𝔫𝔦𝔠𝔬𝔡𝔢𝔫𝔞𝔪𝔢')

        self.writer.clear()
        self.writer.write(NameValuesSerializer, deser)
        self.assertEqual(val, self.writer.copy_buffer())

    def test_pod(self):
        val = self.EXAMPLE_NAMEVALUES
        self.writer.write_bytes(val)
        reader = self._get_reader(pod=True)
        deser = reader.read(NameValuesSerializer)

        self.assertEqual('Title STRING RW DS foo', deser[3])

        self.writer.clear()
        self.writer.write(NameValuesSerializer, deser)
        self.assertEqual(val, self.writer.copy_buffer())

    def test_namevalue_stringify(self):
        test_types = [
            b"Alpha STRING R S 'Twas brillig and the slighy toves/Did gyre and gimble in the wabe",
            b"Beta F32 R S 3.14159",
            b"Gamma S32 R S -12345",
            b"Delta VEC3 R S <1.2, -3.4, 5.6>",
            b"Epsilon U32 R S 12345",
            b"Zeta ASSET R S 041a8591-6f30-42f8-b9f7-7f281351f375",
            b"Eta U64 R S 9223372036854775807"
        ]

        for test in test_types:
            self.writer.clear()
            self.writer.write_bytes(test)
            reader = self._get_reader()
            self.assertEqual(test.decode("utf8"), str(reader.read(NameValueSerializer())))

    def test_namevalues_stringify(self):
        test_list = b"Alpha STRING R S 'Twas brillig and the slighy toves/Did gyre and gimble in the wabe\n" + \
                    b"Beta F32 R S 3.14159\n" + \
                    b"Gamma S32 R S -12345\n" + \
                    b"Delta VEC3 R S <1.2, -3.4, 5.6>\n" + \
                    b"Epsilon U32 R S 12345\n" + \
                    b"Zeta ASSET R S 041a8591-6f30-42f8-b9f7-7f281351f375\n" + \
                    b"Eta U64 R S 9223372036854775807"

        self.writer.clear()
        self.writer.write_bytes(test_list)
        reader = self._get_reader()
        deser = reader.read(NameValuesSerializer)
        self.assertEqual(test_list.decode("utf8"), str(deser))

        # Check that deserializer doesn't raise for any of the fields
        deser.to_dict()


class NumPySerializationTests(BaseSerializationTest):
    def setUp(self) -> None:
        super().setUp()
        self.writer.endianness = "<"

    def test_simple(self):
        quant_spec = se.Vector3U16(0.0, 1.0)
        self.writer.write(quant_spec, Vector3(0, 0.1, 0))
        self.writer.write(quant_spec, Vector3(1, 1, 1))

        reader = self._get_reader()
        np_spec = se.NumPyArray(se.BytesGreedy(), np.dtype(np.uint16), 3)
        np_val = reader.read(np_spec)
        expected_arr = np.array([[0, 6554, 0], [0xFFFF, 0xFFFF, 0xFFFF]], dtype=np.uint16)
        np.testing.assert_array_equal(expected_arr, np_val)

        # Make sure writing the array back works correctly
        orig_buf = self.writer.copy_buffer()
        self.writer.clear()
        self.writer.write(np_spec, expected_arr)
        self.assertEqual(orig_buf, self.writer.copy_buffer())

    def test_quantization(self):
        quant_spec = se.Vector3U16(0.0, 1.0)
        self.writer.write(quant_spec, Vector3(0, 0.1, 0))
        self.writer.write(quant_spec, Vector3(1, 1, 1))

        reader = self._get_reader()
        np_spec = se.QuantizedNumPyArray(se.NumPyArray(se.BytesGreedy(), np.dtype(np.uint16), 3), 0.0, 1.0)
        np_val = reader.read(np_spec)
        expected_arr = np.array([[0, 0.1, 0], [1, 1, 1]], dtype=np.float64)
        np.testing.assert_array_almost_equal(expected_arr, np_val, decimal=5)

        # Make sure writing the array back works correctly
        orig_buf = self.writer.copy_buffer()
        self.writer.clear()
        self.writer.write(np_spec, expected_arr)
        self.assertEqual(orig_buf, self.writer.copy_buffer())


class AnimSerializationTests(BaseSerializationTest):
    SIMPLE_ANIM = b'\x01\x00\x00\x00\x01\x00\x00\x00H\x11\xd1?\x00\x00\x00\x00\x00H\x11\xd1?\x00\x00\x00\x00' \
                  b'\xcd\xccL>\x9a\x99\x99>\x01\x00\x00\x00\x02\x00\x00\x00mNeck\x00\x01\x00\x00\x00\x03\x00' \
                  b'\x00\x00r\n\xff\x7f\xff\x7f\xff\x7f\xfa\xd0\xff\x7f\xff\x7f\xff\x7f\xff\xff\xff\x7f\xff' \
                  b'\x7f\xff\x7f\x00\x00\x00\x00mHead\x00\x01\x00\x00\x00\x14\x00\x00\x00r\n!\x84\xfbz\xab' \
                  b'\x81X\x1f?\x83\xed\x86\xbe\x82\xcb)\xfd\x81\xdc\x86\x08\x83>4\xf9\x7f\xfa~\x92\x82\xb1>' \
                  b'\xdb\x7f\x9d\x80H\x82$I\x82\x81\xad\x89M\x82\x97S\x01\x84\x98\x916\x82\n^L\x86\xfc\x919' \
                  b'\x82}h\xff\x86\xc4\x8c\x93\x82\xefr\x9c\x84\xd7\x85\xe3\x82b}\xb7\x7f\x90\x81\x96\x82\xd5' \
                  b'\x87\xa2~,\x84a\x82H\x92\xd7\x80w\x8aU\x82\xbb\x9c\xa7\x836\x8f4\x82.\xa7\xeb\x84\xa3\x8eX' \
                  b'\x82\xa1\xb1\x9b\x84\x80\x8a\xb6\x82\x14\xbc:\x83\xe8\x85\xf0\x82\x87\xc6^\x813\x83\xd1\x82' \
                  b'\xfa\xd0\xb7\x7f\x90\x81\x96\x82\xff\xff\xb7\x7f\x90\x81\x96\x82\x00\x00\x00\x00\x00\x00\x00\x00'

    def setUp(self) -> None:
        super().setUp()
        self.writer.endianness = "<"

    def test_basic(self):
        self.writer.write_bytes(self.SIMPLE_ANIM)
        spec = se.Dataclass(Animation)
        anim: Animation = self._get_reader().read(spec)
        self.assertEqual(len(anim.joints), 2)
        self.writer.clear()
        self.writer.write(spec, anim)
        self.assertEqual(self.SIMPLE_ANIM, self.writer.copy_buffer())

    def test_elided_fields_allowed(self):
        spec = se.Dataclass(Animation)
        # Should use the defaults for the unspecified fields
        anim = Animation(
            major_version=1,
            minor_version=0,
            base_priority=5,
            duration=1.0,
            loop_out_point=1.0,
        )
        self.assertIsNotNone(anim)
        self.assertEqual(len(anim.joints), 0)
        anim.joints["mPelvis"] = Joint(priority=5, rot_keyframes=[
            RotKeyframe(time=0.0, rot=Quaternion(0, 0, 0, 1))
        ])
        self.assertEqual(len(anim.joints), 1)
        self.writer.write(spec, anim)
        reader = self._get_reader()
        deser = reader.read(spec)
        self.assertEqual(anim, deser)

    def test_dict_context_allowed(self):
        spec = se.Dataclass(RotKeyframe)
        kf = RotKeyframe(time=0.0, rot=Quaternion())
        ctx = se.ParseContext({"major_version": 1, "minor_version": 0, "duration": 5.0})
        self.writer.write(spec, kf, ctx=ctx)
        # U16 for time, (x, y, z)
        self.assertEqual(se.U16.calc_size() * 4, len(self.writer))

        self.writer.clear()
        # Shouldn't need duration for v0.1
        ctx = se.ParseContext({"major_version": 0, "minor_version": 1})
        self.writer.write(spec, kf, ctx=ctx)
        # F32 for time, (x, y, z)
        self.assertEqual(se.F32.calc_size() * 4, len(self.writer))


class SubfieldSerializationTests(BaseSerializationTest):
    def test_enum(self):
        class FooEnum(enum.IntEnum):
            FOO = 1
            BAR = 2

        ser = se.IntEnumSubfieldSerializer(FooEnum)

        self.assertEqual(ser.deserialize(None, 1, pod=False), FooEnum.FOO)
        self.assertEqual(ser.deserialize(None, 1, pod=True), "FOO")
        self.assertEqual(ser.deserialize(None, 3, pod=True), se.UNSERIALIZABLE)
        self.assertEqual(ser.deserialize(None, 3, pod=False), 3)

        self.assertEqual(ser.serialize(None, "FOO"), 1)
        self.assertEqual(ser.serialize(None, FooEnum.FOO), 1)
        self.assertEqual(ser.serialize(None, 3), 3)

    def test_flags(self):
        class FooFlags(enum.IntFlag):
            FOO = 1
            BAR = 1 << 1

        ser = se.IntFlagSubfieldSerializer(FooFlags)

        self.assertEqual(ser.deserialize(None, 1, pod=False), FooFlags.FOO)
        self.assertEqual(ser.deserialize(None, 3, pod=False), FooFlags.FOO | FooFlags.BAR)
        self.assertEqual(ser.deserialize(None, 1, pod=True), ("FOO",))
        self.assertEqual(ser.deserialize(None, 3, pod=True), ("FOO", "BAR"))
        self.assertEqual(ser.deserialize(None, 7, pod=True), ("FOO", "BAR", 4))

        self.assertEqual(ser.serialize(None, ()), 0)
        self.assertEqual(ser.serialize(None, 0), 0)
        self.assertEqual(ser.serialize(None, ("FOO", "BAR")), 3)
        self.assertEqual(ser.serialize(None, FooFlags.FOO), 1)
        self.assertEqual(ser.serialize(None, 3), 3)
        self.assertEqual(ser.serialize(None, 7), 7)



import math
import unittest

import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.datatypes import UUID, Vector3
from hippolyzer.lib.base.message.message_formatting import HumanMessageSerializer
from hippolyzer.lib.base.templates import TextureEntrySubfieldSerializer, TEFaceBitfield, TextureEntryCollection, \
    PackedTERotation, TextureEntry

EXAMPLE_TE = b'\x89UgG$\xcbC\xed\x92\x0bG\xca\xed\x15F_\x08\xca*\x98:\x18\x02,\r\xf4\x1e\xc6\xf5\x91\x01]\x83\x014' \
             b'\x00\x90i+\x10\x80\xa1\xaa\xa2g\x11o\xa8]\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x80?\x00\x00\x00\x80?' \
             b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
             b'\x00\x00\x00\x00\x00\x00\x00'


class TemplateTests(unittest.TestCase):
    def test_te_round_trips(self):
        deserialized = TextureEntrySubfieldSerializer.deserialize(None, EXAMPLE_TE)
        serialized = TextureEntrySubfieldSerializer.serialize(None, deserialized)
        self.assertEqual(EXAMPLE_TE, serialized)

    def test_realize_te(self):
        deserialized: TextureEntryCollection = TextureEntrySubfieldSerializer.deserialize(None, EXAMPLE_TE)
        realized = deserialized.realize(4)
        self.assertEqual(UUID('ca2a983a-1802-2c0d-f41e-c6f591015d83'), realized[3].Textures)
        self.assertEqual(UUID('89556747-24cb-43ed-920b-47caed15465f'), realized[1].Textures)
        with self.assertRaises(ValueError):
            deserialized.realize(3)

    def test_tecollection_from_tes(self):
        deserialized: TextureEntryCollection = TextureEntrySubfieldSerializer.deserialize(None, EXAMPLE_TE)
        # The TE collection should re-serialize to the same collection when split up and regrouped
        self.assertEqual(deserialized, TextureEntryCollection.from_tes(deserialized.realize(4)))

    def test_face_bitfield_round_trips(self):
        test_val = b"\x81\x03"
        reader = se.BufferReader("!", test_val)
        deserialized = reader.read(TEFaceBitfield)

        writer = se.BufferWriter("!")
        writer.write(TEFaceBitfield, deserialized)
        self.assertEqual(writer.copy_buffer(), test_val)

    def test_can_use_templated_pod_message(self):
        pod_te = {
            'Textures': {
                None: '89556747-24cb-43ed-920b-47caed15465f',
                (3,): 'ca2a983a-1802-2c0d-f41e-c6f591015d83',
                (0,): '34009069-2b10-80a1-aaa2-67116fa85dc6'
            },
            'Color': {None: b'\xff\xff\xff\xff'},
            'ScalesS': {None: 1.0},
            'ScalesT': {None: 1.0},
            'OffsetsS': {None: 0.0},
            'OffsetsT': {None: 0.0},
            'Rotation': {None: 0.0},
            'BasicMaterials': {None: {'Bump': 0, 'FullBright': False, 'Shiny': 'OFF'}},
            'MediaFlags': {None: {'WebPage': False, 'TexGen': 'DEFAULT', '_Unused': 0}}, 'Glow': {None: 0},
            'Materials': {None: '00000000-0000-0000-0000-000000000000'},
        }
        msg = HumanMessageSerializer.from_human_string(f"""
        OUT ObjectImage
        [AgentData]
          AgentID = {UUID()}
          SessionID = {UUID()}
        [ObjectData]
          ObjectLocalID = 700966
          MediaURL = b''
          TextureEntry =| {repr(pod_te)}
        """)
        # Make sure from/to_human_string doesn't change meaning
        str_msg = HumanMessageSerializer.to_human_string(msg, beautify=True)
        msg = HumanMessageSerializer.from_human_string(str_msg)
        spec = msg["ObjectData"][0].get_serializer("TextureEntry")
        data_field = msg["ObjectData"]["TextureEntry"]
        # Serialization order and format should match indra's exactly
        self.assertEqual(EXAMPLE_TE, data_field)
        deser = spec.deserialize(None, data_field, pod=True)
        self.assertEqual(pod_te, deser)

    def test_textureentry_defaults(self):
        te = TextureEntryCollection()
        self.assertEqual(UUID('89556747-24cb-43ed-920b-47caed15465f'), te.Textures[None])

    def test_textureentry_rotation_packing(self):
        writer = se.BufferWriter("!")
        writer.write(PackedTERotation(), math.pi * 2)
        # fmod() makes this loop back around to 0
        self.assertEqual(b"\x00\x00", writer.copy_buffer())
        writer.clear()

        writer.write(PackedTERotation(), -math.pi * 2)
        # fmod() makes this loop back around to 0
        self.assertEqual(b"\x00\x00", writer.copy_buffer())
        writer.clear()

        writer.write(PackedTERotation(), 0)
        self.assertEqual(b"\x00\x00", writer.copy_buffer())
        writer.clear()

        # These both map to -32768 because of overflow in the positive case
        # that isn't caught by exact equality to math.pi * 2
        writer.write(PackedTERotation(), math.pi * 1.999999)
        self.assertEqual(b"\x80\x00", writer.copy_buffer())
        writer.clear()

        writer.write(PackedTERotation(), math.pi * -1.999999)
        self.assertEqual(b"\x80\x00", writer.copy_buffer())
        writer.clear()

    def test_textureentry_rotation_unpacking(self):
        reader = se.BufferReader("!", b"\x00\x00")
        self.assertEqual(0, reader.read(PackedTERotation()))

        reader = se.BufferReader("!", b"\x80\x00")
        self.assertEqual(-math.pi * 2, reader.read(PackedTERotation()))

        # This quantization method does not allow for any representation of
        # F_TWO_PI itself, just a value slightly below it! The float representation
        # is ever so slightly different from the C++ version, but it should still
        # round-trip correctly.
        reader = se.BufferReader("!", b"\x7f\xff")
        self.assertEqual(6.282993559581101, reader.read(PackedTERotation()))

        writer = se.BufferWriter("!")
        writer.write(PackedTERotation(), 6.282993559581101)
        self.assertEqual(b"\x7f\xff", writer.copy_buffer())

    def test_textureentry_st_to_uv_coords(self):
        te = TextureEntry(ScalesS=0.5, ScalesT=0.5, OffsetsS=-0.25, OffsetsT=0.25, Rotation=math.pi / 2)
        self.assertEqual(Vector3(0.25, 0.75), te.st_to_uv(Vector3(0.5, 0.5)))



import collections

from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.addon_utils import BaseAddon, GlobalProperty
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class PacketStatsAddon(BaseAddon):
    packet_stats: collections.Counter = GlobalProperty(collections.Counter)

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        self.packet_stats[message.name] += 1

    @handle_command()
    async def print_packet_stats(self, _session: Session, _region: ProxiedRegion):
        print(self.packet_stats.most_common(10))


addons = [PacketStatsAddon()]



"""
Speed up outbound object inventory listing requests
by 20x at the cost of potentially failing to request some due to
dropped packets.

Useful for builders working on objects with very large inventories that
change very often.

Object Inventory transfers use the Xfer system. Xfers have their own,
terrible reliability system that probably pre-dates LLUDP reliability.
Each packet has to be ACKed before the far end will send the next packet.
Each packet can be around 1200 bytes and will fit 1.5 inventory items worth of data.

Let's say your sim ping is 100 ms. Because each packet needs to be ACKed
before the next will be sent, it'll take around `num_items * 100 / 1.5`
milliseconds before you receive the full inventory list of an object.
That means for an object with 300 items, it'll take about 20 seconds
for you to download the full inventory, and those downloads are triggered
every time the inventory is changed.

By faking ACKs for packets we haven't received yet, we can trick the server
into sending us packets much faster than it would otherwise. The only problem
is that if an inbound SendXferPacket gets lost after we faked an ACK for it,
we have no way to re-request it. The Xfer will just fail. The viewer will also
drop any out-of-order xfer packets, so packet re-ordering is a problem.

To deal with that, the proxy attempts its own Xfers using all the chunks
from the previous attempts before sending a final, reconstructed Xfer
to the viewer.
"""

import asyncio
from typing import *

from hippolyzer.lib.base.templates import XferFilePath
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session
from hippolyzer.lib.base.xfer_manager import Xfer


class TurboObjectInventoryAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.direction != Direction.OUT:
            return
        if message.name != "RequestTaskInventory":
            return

        self._schedule_task(self._proxy_task_inventory_request(region, message.take()))
        return True

    async def _proxy_task_inventory_request(
            self,
            region: ProxiedRegion,
            request_msg: Message
    ):
        # Keep around a dict of chunks we saw previously in case we have to restart
        # an Xfer due to missing chunks. We don't expect chunks to change across Xfers
        # so this can be used to recover from dropped SendXferPackets in subsequent attempts
        existing_chunks: Dict[int, bytes] = {}
        for i in range(3):
            # Any previous requests will have triggered a delete of the inventory file
            # by marking it complete on the server-side. Re-send our RequestTaskInventory
            # To make sure there's a fresh copy.
            region.circuit.send(request_msg.take())
            inv_message = await region.message_handler.wait_for(('ReplyTaskInventory',), timeout=5.0)
            # No task inventory, send the reply as-is
            file_name = inv_message["InventoryData"]["Filename"]
            if not file_name:
                region.circuit.send(inv_message)
                return

            xfer = region.xfer_manager.request(
                file_name=file_name,
                file_path=XferFilePath.CACHE,
                turbo=True,
            )
            xfer.chunks.update(existing_chunks)
            try:
                await xfer
            except asyncio.TimeoutError:
                # We likely failed the request due to missing chunks, store
                # the chunks that we _did_ get for the next attempt.
                existing_chunks.update(xfer.chunks)
                continue

            # Send the original ReplyTaskInventory to the viewer so it knows the file is ready
            region.circuit.send(inv_message)
            proxied_xfer = Xfer(data=xfer.reassemble_chunks())

            # Wait for the viewer to request the inventory file
            await region.xfer_manager.serve_inbound_xfer_request(
                xfer=proxied_xfer,
                request_predicate=lambda x: x["XferID"]["Filename"] == file_name,
                # indra's XferManager throttles confirms, so even local transfers will be
                # slow if we wait for confirmation.
                wait_for_confirm=False,
            )
            return
        raise asyncio.TimeoutError("Failed to get inventory after 3 tries")


addons = [TurboObjectInventoryAddon()]



"""
A simple client that just says hello to people
"""

import asyncio
import pprint
from contextlib import aclosing
import os

from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.templates import ChatType, ChatSourceType
from hippolyzer.lib.client.hippo_client import HippoClient


async def amain():
    client = HippoClient()

    async def _respond_to_chat(message: Message):
        if message["ChatData"]["SourceID"] == client.session.agent_id:
            return
        if message["ChatData"]["SourceType"] != ChatSourceType.AGENT:
            return
        if "hello" not in message["ChatData"]["Message"].lower():
            return
        await client.send_chat(f'Hello {message["ChatData"]["FromName"]}!', chat_type=ChatType.SHOUT)

    async with aclosing(client):
        await client.login(
            username=os.environ["HIPPO_USERNAME"],
            password=os.environ["HIPPO_PASSWORD"],
            start_location=os.environ.get("HIPPO_START_LOCATION", "last"),
        )
        print("I'm here")

        # Wait until we have details about parcels and print them
        await client.main_region.parcel_manager.parcels_downloaded.wait()
        pprint.pprint(client.main_region.parcel_manager.parcels)

        await client.send_chat("Hello World!", chat_type=ChatType.SHOUT)
        client.session.message_handler.subscribe("ChatFromSimulator", _respond_to_chat)
        # Example of how to work with caps
        async with client.main_caps_client.get("SimulatorFeatures") as features_resp:
            print("Features:", await features_resp.read_llsd())

        while True:
            try:
                await asyncio.sleep(0.001)
            except (KeyboardInterrupt, asyncio.CancelledError):
                await client.send_chat("Goodbye World!", chat_type=ChatType.SHOUT)
                return

if __name__ == "__main__":
    asyncio.run(amain())



import unittest

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.client.inventory_manager import InventoryManager
from tests.client import MockClientRegion

CREATE_FOLDER_PAYLOAD = {
    '_base_uri': 'slcap://InventoryAPIv3',
    '_created_categories': [
        UUID(int=2),
    ],
    '_created_items': [],
    '_embedded': {
        'categories': {
            f'{UUID(int=2)}': {
                '_embedded': {'categories': {}, 'items': {}, 'links': {}},
                '_links': {
                    'parent': {'href': f'/category/{UUID(int=1)}'},
                    'self': {'href': f'/category/{UUID(int=2)}'}
                },
                'agent_id': f'{UUID(int=9)}',
                'category_id': f'{UUID(int=2)}',
                'name': 'New Folder',
                'parent_id': f'{UUID(int=1)}',
                'type_default': -1,
                'version': 1
            }
        },
        'items': {}, 'links': {}
    },
    '_links': {
        'categories': {'href': f'/category/{UUID(int=1)}/categories'},
        'category': {'href': f'/category/{UUID(int=1)}', 'name': 'self'},
        'children': {'href': f'/category/{UUID(int=1)}/children'},
        'items': {'href': f'/category/{UUID(int=1)}/items'},
        'links': {'href': f'/category/{UUID(int=1)}/links'},
        'parent': {'href': '/category/00000000-0000-0000-0000-000000000000'},
        'self': {'href': f'/category/{UUID(int=1)}/children'}
    },
    '_updated_category_versions': {str(UUID(int=1)): 27},
    'agent_id': UUID(int=9),
    'category_id': UUID(int=1),
    'name': 'My Inventory',
    'parent_id': UUID.ZERO,
    'type_default': 8,
    'version': 27,
}


class TestParcelOverlay(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.region = MockClientRegion()
        self.session = self.region.session()
        self.inv_manager = InventoryManager(self.session)
        self.model = self.inv_manager.model
        self.handler = self.region.message_handler

    def test_create_folder_response(self):
        self.inv_manager.process_aisv3_response(CREATE_FOLDER_PAYLOAD)
        self.assertIsNotNone(self.model.get_category(UUID(int=1)))
        self.assertIsNotNone(self.model.get_category(UUID(int=2)))



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import pickle
import unittest
import uuid

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import *


class TestDatatypes(unittest.TestCase):

    def tearDown(self):
        pass

    def setUp(self):
        pass

    def test_vector3_new(self):
        vector = Vector3()
        self.assertEqual(vector, (0.0, 0.0, 0.0))

    def test_vector3_from_xyz(self):
        vector = Vector3(X=128.0, Y=128.0, Z=22.0)
        self.assertEqual(vector.X, 128.0)
        self.assertEqual(vector.Y, 128.0)
        self.assertEqual(vector.Z, 22.0)

    def test_vector3_addition(self):
        vector = Vector3(1, 2, 3)
        vector2 = Vector3(0, 1, 2)
        self.assertEqual(vector + vector2, (1.0, 3.0, 5.0))

    def test_vector3_subtraction(self):
        vector = Vector3(1, 2, 3)
        vector2 = Vector3(0, 1, 2)
        self.assertEqual(vector - vector2, (1.0, 1.0, 1.0))

    def test_vector3_comparison(self):
        vector = Vector3(1, 2, 3)
        self.assertTrue(vector >= (0, 1, 2))
        self.assertTrue(vector < (4, 4, 4))

    def test_vector3_iteration(self):
        self.assertEqual((1, 2, 3), tuple(x for x in Vector3(1, 2, 3)))

    def test_vector3_str(self):
        self.assertEqual(str(Vector3(0, 0, 0)), "<0.0, 0.0, 0.0>")

    def test_vector3_repr(self):
        self.assertEqual(repr(Vector3(0, 0, 0)), "Vector3(0.0, 0.0, 0.0)")

    def test_vector3_parse(self):
        self.assertEqual(Vector3(0, 1, 0), Vector3.parse("<0, 1, 0>"))

    def test_quaternion_new(self):
        quat = Quaternion()
        self.assertEqual(quat, (0.0, 0.0, 0.0, 1.0))

    def test_quaternion_from_xyzw(self):
        quat = Quaternion(X=128.0, Y=128.0, Z=22.0)
        self.assertEqual(quat, (128.0, 128.0, 22.0, 0.0))

    def test_quaternion_euler_roundtrip(self):
        orig_vec = Vector3(0.0, -1.0, 2.0)
        quat = Quaternion.from_euler(*orig_vec)
        for orig_comp, new_comp in zip(orig_vec, quat.to_euler()):
            self.assertAlmostEqual(orig_comp, new_comp)

    def test_quaternion_transformations(self):
        quat = Quaternion(0.4034226801113349, -0.2590347239999257, 0.7384602626041288, 0.4741598817790379)
        expected_trans = (0.4741598817790379, 0.4034226801113349, -0.2590347239999257, 0.7384602626041288)
        trans_quat = quat.to_transformations()
        self.assertSequenceEqual(expected_trans, trans_quat)
        new_quat = Quaternion.from_transformations(trans_quat)
        self.assertEqual(quat, new_quat)

    def test_uuid_from_bytes(self):
        tmp_uuid = uuid.UUID('2b7f7a6e-32c5-dbfd-e2c7-926d1a9f0aca')
        tmp_uuid2 = uuid.UUID('1dd5efe2-faaf-1864-5ac9-bc61c5d8d7ea')

        test_uuid = UUID(bytes=tmp_uuid.bytes)
        test_uuid2 = UUID(bytes=tmp_uuid2.bytes)

        self.assertEqual(test_uuid, uuid.UUID('2b7f7a6e-32c5-dbfd-e2c7-926d1a9f0aca'))
        self.assertEqual(test_uuid2, uuid.UUID('1dd5efe2-faaf-1864-5ac9-bc61c5d8d7ea'))

    def test_uuid_get_bytes(self):
        test_uuid = UUID(val='12345678-1234-1234-1234-123456789012')
        self.assertEqual(test_uuid.bytes, uuid.UUID('12345678-1234-1234-1234-123456789012').bytes)

    def test_uuid_from_string(self):
        uuid_string = '2b7f7a6e-32c5-dbfd-e2c7-926d1a9f0aca'
        self.assertEqual(UUID(val=uuid_string), uuid.UUID('2b7f7a6e-32c5-dbfd-e2c7-926d1a9f0aca'))

    def test_uuid_combine(self):
        first = UUID('00000000-0000-0000-0000-000000000000')
        second = uuid.UUID('00000000-0100-0000-0000-000000000000')
        combined = UUID.combine(first, second)
        self.assertEqual(combined, UUID('8c4b3740-2538-dc5d-1ec2-6cb50e2c348b'))

    def test_uuid_equality(self):
        first = UUID('00000000-0000-0000-0000-000000000000')
        second = uuid.UUID('00000000-0000-0000-0000-000000000000')
        self.assertEqual(second, first)

    def test_uuid_hash_equality(self):
        first = UUID('00000000-0000-0000-0000-000000000000')
        second = uuid.UUID('00000000-0000-0000-0000-000000000000')
        self.assertEqual(hash(first), hash(second))

    def test_uuid_xor(self):
        first = UUID('00000000-0000-0000-0000-000000000101')
        second = uuid.UUID('00000000-0000-0000-0000-000000000110')
        self.assertEqual(UUID('00000000-0000-0000-0000-000000000011'), first ^ second)

    def test_pickleable(self):
        orig = Vector3(1, 2, 3)
        val = pickle.loads(pickle.dumps(orig, protocol=pickle.HIGHEST_PROTOCOL))
        self.assertEqual(orig, val)

    def test_tuplecoord_llsd_serialization(self):
        orig = Vector3(1, 2, 3)
        val = llsd.parse_xml(llsd.format_xml(orig))
        self.assertIsInstance(val, list)
        self.assertEqual(orig, tuple(val))

    def test_uuid_llsd_serialization(self):
        orig = UUID.random()
        val = llsd.parse_binary(llsd.format_binary(orig))
        self.assertIsInstance(val, UUID)
        self.assertEqual(orig, val)

    def test_str_llsd_serialization(self):
        self.assertEqual(b"'foo\\nbar'", llsd.format_notation("foo\nbar"))

    def test_int_enum_llsd_serialization(self):
        class SomeIntEnum(IntEnum):
            FOO = 4

        orig = SomeIntEnum.FOO
        val = llsd.parse_xml(llsd.format_xml(orig))
        self.assertIsInstance(val, int)
        self.assertEqual(orig, val)

    def test_jank_stringy_bytes(self):
        val = JankStringyBytes(b"foo\x00")
        self.assertTrue("o" in val)
        self.assertTrue(b"o" in val)
        self.assertFalse(b"z" in val)
        self.assertFalse("z" in val)
        self.assertEqual("foo", val)
        self.assertEqual(b"foo\x00", val)
        self.assertNotEqual(b"foo", val)
        self.assertEqual(b"foo", JankStringyBytes(b"foo"))
        self.assertEqual("foo", JankStringyBytes(b"foo"))
        self.assertFalse(JankStringyBytes(b""))
        self.assertFalse(JankStringyBytes(b"\x00"))
        self.assertTrue(JankStringyBytes(b"\x01"))



"""
Example of custom meta tags, useful for complex expressions that wouldn't work
well in the message log filter language.

Tags messages where someone said "hello", and record who they said hello to.

If you said "hello Someone", that message would be shown in the log pane when
filtering with `Meta.Greeted == "Someone"` or just `Meta.Greeted` to match any
message with a greeting.
"""

from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class CustomMetaExampleAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if not message.name.startswith("ChatFrom"):
            return

        chat = message["ChatData"]["Message"]
        if not chat:
            return

        if chat.lower().startswith("hello "):
            message.meta["Greeted"] = chat.split(" ", 1)[1]


addons = [CustomMetaExampleAddon()]



"""
Debug performance issues in the proxy
/524 start_profiling
/524 stop_profiling
"""

import cProfile
from typing import *

from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


class ProfilingAddon(BaseAddon):
    def __init__(self):
        # We don't want this to surive module reloads so it can be an
        # instance attribute rather than on session_manager.addon_ctx
        self.profile: Optional[cProfile.Profile] = None

    def handle_unload(self, session_manager: SessionManager):
        if self.profile is not None:
            self.profile.disable()
            self.profile = None

    @handle_command()
    async def start_profiling(self, _session: Session, _region: ProxiedRegion):
        """Start a cProfile session"""
        if self.profile is not None:
            self.profile.disable()
        self.profile = cProfile.Profile()
        self.profile.enable()
        print("Started profiling")

    @handle_command()
    async def stop_profiling(self, _session: Session, _region: ProxiedRegion):
        """Stop profiling and save to file"""
        if self.profile is None:
            return
        self.profile.disable()
        profile = self.profile
        self.profile = None

        print("Finished profiling")
        profile_path = await AddonManager.UI.save_file(caption="Save Profile")
        if profile_path:
            profile.dump_stats(profile_path)


addons = [ProfilingAddon()]



"""
Example of how to request a Transfer
"""
from typing import *

from hippolyzer.lib.base.inventory import InventoryModel, InventoryItem
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.templates import (
    AssetType,
    EstateAssetType,
    TransferRequestParamsSimEstate,
    TransferRequestParamsSimInvItem,
    TransferSourceType,
    XferFilePath,
)
from hippolyzer.lib.proxy.addon_utils import BaseAddon, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class TransferExampleAddon(BaseAddon):
    @handle_command()
    async def get_covenant(self, _session: Session, region: ProxiedRegion):
        """Get the current region's covenent"""
        transfer = await region.transfer_manager.request(
            source_type=TransferSourceType.SIM_ESTATE,
            params=TransferRequestParamsSimEstate(
                EstateAssetType=EstateAssetType.COVENANT,
            ),
        )
        show_message(transfer.reassemble_chunks().decode("utf8"))

    @handle_command()
    async def get_first_script(self, session: Session, region: ProxiedRegion):
        """Get the contents of the first script in the selected object"""
        # Ask for the object inventory so we can find a script
        region.circuit.send(Message(
            'RequestTaskInventory',
            Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
            Block('InventoryData', LocalID=session.selected.object_local),
        ))
        inv_message = await region.message_handler.wait_for(('ReplyTaskInventory',), timeout=5.0)

        # Xfer the inventory file and look for a script
        xfer = await region.xfer_manager.request(
            file_name=inv_message["InventoryData"]["Filename"], file_path=XferFilePath.CACHE)
        inv_model = InventoryModel.from_bytes(xfer.reassemble_chunks())
        first_script: Optional[InventoryItem] = None
        for item in inv_model.all_items:
            if item.type == "lsltext":
                first_script = item
        if not first_script:
            show_message("No scripts in object?")
            return

        # Ask for the actual script contents
        transfer = await region.transfer_manager.request(
            source_type=TransferSourceType.SIM_INV_ITEM,
            params=TransferRequestParamsSimInvItem(
                OwnerID=first_script.permissions.owner_id,
                TaskID=inv_model.root.node_id,
                ItemID=first_script.item_id,
                AssetType=AssetType.LSL_TEXT,
            ),
        )
        show_message(transfer.reassemble_chunks().decode("utf8"))


addons = [TransferExampleAddon()]



import asyncio
import copy
import unittest
import xmlrpc.client
from typing import Tuple, Optional

import aioresponses

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.circuit import Circuit
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.message.msgtypes import PacketFlags
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.network.transport import AbstractUDPTransport, UDPPacket, Direction
from hippolyzer.lib.base.test_utils import MockTransport, MockConnectionHolder, soon
from hippolyzer.lib.client.hippo_client import HippoClient, HippoClientProtocol


class MockServer(MockConnectionHolder):
    def __init__(self, circuit, message_handler):
        super().__init__(circuit, message_handler)
        self.deserializer = UDPMessageDeserializer()
        self.protocol: Optional[HippoClientProtocol] = None

    def process_inbound(self, packet: UDPPacket):
        """Process a packet that the client sent to us"""
        message = self.deserializer.deserialize(packet.data)
        message.direction = Direction.IN
        if message.reliable:
            self.circuit.send_acks((message.packet_id,))
        self.circuit.collect_acks(message)
        if message.name != "PacketAck":
            self.message_handler.handle(message)


class PacketForwardingTransport(MockTransport):
    def __init__(self):
        super().__init__()
        self.protocol: Optional[HippoClientProtocol] = None

    def send_packet(self, packet: UDPPacket):
        super().send_packet(packet)
        self.protocol.datagram_received(packet.data, packet.src_addr)


class MockServerTransport(MockTransport):
    """Used for the client to send packets out"""
    def __init__(self, server: MockServer):
        super().__init__()
        self._server = server

    def send_packet(self, packet: UDPPacket) -> None:
        super().send_packet(packet)
        # Directly pass the packet to the server
        packet = copy.copy(packet)
        packet.direction = Direction.IN
        # Delay calling so the client can do its ACK bookkeeping first
        asyncio.get_event_loop().call_soon(lambda: self._server.process_inbound(packet))


class MockHippoClient(HippoClient):
    def __init__(self, server: MockServer):
        super().__init__()
        self.server = server

    async def _create_transport(self) -> Tuple[AbstractUDPTransport, HippoClientProtocol]:
        protocol = HippoClientProtocol(self.session)
        # TODO: This isn't great, but whatever.
        self.server.circuit.transport.protocol = protocol
        return MockServerTransport(self.server), protocol


class TestHippoClient(unittest.IsolatedAsyncioTestCase):
    FAKE_LOGIN_URI = "http://127.0.0.1:1/login.cgi"
    FAKE_LOGIN_RESP = {
        "session_id": str(UUID(int=1)),
        "secure_session_id": str(UUID(int=2)),
        "agent_id": str(UUID(int=3)),
        "circuit_code": 123,
        "sim_ip": "127.0.0.1",
        "sim_port": 2,
        "region_x": 0,
        "region_y": 123,
        "seed_capability": "https://127.0.0.1:4/foo",
        "inventory-skeleton": [
            {'name': 'My Inventory', 'folder_id': str(UUID(int=4)),
             'parent_id': '00000000-0000-0000-0000-000000000000', 'type_default': 8, 'version': 200}
        ]
    }
    FAKE_SEED_RESP = {
        "EventQueueGet": "https://127.0.0.1:5/",
    }
    FAKE_EQ_RESP = {
        "id": 1,
        "events": [
            {"message": "ViewerFrozenMessage", "body": {"FrozenData": [{"Data": False}]}},
            {"message": "NotTemplated", "body": {"foo": {"bar": True}}},
        ],
    }

    async def asyncSetUp(self):
        self.server_handler: MessageHandler[Message, str] = MessageHandler()
        self.server_transport = PacketForwardingTransport()
        self.server_circuit = Circuit(("127.0.0.1", 2), ("127.0.0.1", 99), self.server_transport)
        self.server = MockServer(self.server_circuit, self.server_handler)
        self.aio_mock = aioresponses.aioresponses()
        self.aio_mock.start()
        self.aio_mock.post(
            self.FAKE_LOGIN_URI,
            body=xmlrpc.client.dumps((self.FAKE_LOGIN_RESP,), None, True)
        )
        self.aio_mock.post(self.FAKE_LOGIN_RESP['seed_capability'], body=llsd.format_xml(self.FAKE_SEED_RESP))
        self.aio_mock.post(self.FAKE_SEED_RESP['EventQueueGet'], body=llsd.format_xml(self.FAKE_EQ_RESP), repeat=True)
        self.client = MockHippoClient(self.server)

    async def asyncTearDown(self):
        try:
            await self.client.aclose()
        finally:
            self.aio_mock.stop()

    async def _log_client_in(self, client: MockHippoClient):
        login_task = asyncio.create_task(client.login("foo", "bar", login_uri=self.FAKE_LOGIN_URI))
        with self.server_handler.subscribe_async(
                ("*",),
        ) as get_msg:
            assert (await soon(get_msg())).name == "UseCircuitCode"
            assert (await soon(get_msg())).name == "CompleteAgentMovement"
            self.server.circuit.send(Message(
                'RegionHandshake',
                Block('RegionInfo', fill_missing=True),
                Block('RegionInfo2', fill_missing=True),
                Block('RegionInfo3', fill_missing=True),
                Block('RegionInfo4', fill_missing=True),
            ))
            assert (await soon(get_msg())).name == "RegionHandshakeReply"
            assert (await soon(get_msg())).name == "AgentThrottle"
            await login_task

    async def test_login(self):
        await self._log_client_in(self.client)
        with self.server_handler.subscribe_async(
                ("*",),
        ) as get_msg:
            self.client.logout()
            assert (await soon(get_msg())).name == "LogoutRequest"

    async def test_eq(self):
        await self._log_client_in(self.client)
        with self.client.session.message_handler.subscribe_async(
                ("ViewerFrozenMessage", "NotTemplated"),
        ) as get_msg:
            assert (await soon(get_msg())).name == "ViewerFrozenMessage"
            msg = await soon(get_msg())
            assert msg.name == "NotTemplated"
            assert msg["EventData"]["foo"]["bar"] == 1

    async def test_inventory_manager(self):
        await self._log_client_in(self.client)
        self.assertEqual(self.client.session.inventory.model.root.node_id, UUID(int=4))

    async def test_resend_suppression(self):
        """Make sure the client only handles the first seen copy of a reliable message"""
        await self._log_client_in(self.client)
        with self.client.session.message_handler.subscribe_async(
                ("ChatFromSimulator", "AgentDataUpdate"),
        ) as get_msg:
            msg = Message("ChatFromSimulator", Block("ChatData", fill_missing=True))
            msg.send_flags |= PacketFlags.RELIABLE

            # Fake re-sending the message
            packet = self.server_circuit.send(msg)
            self.server_transport.send_packet(packet)

            self.server_circuit.send(Message("AgentDataUpdate", Block("AgentData", fill_missing=True)))
            assert (await soon(get_msg())).name == "ChatFromSimulator"
            assert (await soon(get_msg())).name == "AgentDataUpdate"



"""Block potentially bad things"""
from hippolyzer.lib.base.templates import IMDialogType, XferFilePath
from hippolyzer.lib.proxy.addon_utils import BaseAddon, show_message
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session

SUSPICIOUS_PACKETS = {
    "TransferRequest",
    "UUIDNameRequest",
    "UUIDGroupNameRequest",
    "OpenCircuit",
    "AddCircuitCode",
}
REGULAR_IM_DIALOGS = (IMDialogType.TYPING_STOP, IMDialogType.TYPING_STOP, IMDialogType.NOTHING_SPECIAL)


class ShieldAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.direction != Direction.IN:
            return
        if message.name in SUSPICIOUS_PACKETS:
            show_message(f"Blocked suspicious {message.name} packet")
            region.circuit.drop_message(message)
            return True
        if message.name == "ImprovedInstantMessage":
            msg_block = message["MessageBlock"][0]
            if msg_block["Dialog"] not in REGULAR_IM_DIALOGS:
                return
            from_agent = message["AgentData"]["AgentID"]
            if from_agent == session.agent_id:
                expected_id = from_agent
            else:
                expected_id = from_agent ^ session.agent_id
            msg_block["ID"] = expected_id
        if message.name == "RequestXfer":
            xfer_block = message["XferID"][0]
            # Don't allow Xfers for files, only assets
            if xfer_block["FilePath"] != XferFilePath.NONE or xfer_block["Filename"]:
                show_message(f"Blocked suspicious {message.name} packet")
                region.circuit.drop_message(message)
                return True


addons = [ShieldAddon()]



"""
Body horror local animation mutator

Demonstrates programmatic modification / generation of animations

It will make you look absurd, obscene.
"""
import copy

import mitmproxy.http

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.llanim import Animation
from hippolyzer.lib.proxy.addon_utils import AssetAliasTracker, BaseAddon, GlobalProperty
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager
from hippolyzer.lib.base.vfs import STATIC_VFS


JOINT_REPLS = {
    "Left": "Right",
    "Right": "Left",
    "LEFT": "RIGHT",
    "RIGHT": "LEFT",
}


def _change_joint_name(joint_name: str):
    for orig, repl in JOINT_REPLS.items():
        if orig in joint_name:
            return joint_name.replace(orig, repl)
    return joint_name


def _mutate_anim_bytes(anim_bytes: bytes):
    anim = Animation.from_bytes(anim_bytes)
    new_joints = {}
    for name, joint in anim.joints.items():
        new_joints[_change_joint_name(name)] = joint
    anim.joints = new_joints
    for constraint in anim.constraints:
        constraint.source_volume = _change_joint_name(constraint.source_volume)
        constraint.target_volume = _change_joint_name(constraint.target_volume)
    return anim.to_bytes()


class HorrorAnimatorAddon(BaseAddon):
    horror_anim_tracker: AssetAliasTracker = GlobalProperty(AssetAliasTracker)

    def handle_init(self, session_manager: SessionManager):
        # We've reloaded, so make sure assets get new aliases
        self.horror_anim_tracker.invalidate_aliases()

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        tracker = self.horror_anim_tracker

        if message.name == "AvatarAnimation":
            # Only do this for the current user
            if message["Sender"]["ID"] != session.agent_id:
                return
            # Replace inbound anim IDs with alias IDs so we can force a cache
            # miss and replace the contents
            for block in message["AnimationList"][:]:
                anim_id = block["AnimID"]
                # Many of the anims in the static VFS have special meanings and the viewer
                # does different things based on the presence or absence of their IDs
                # in the motion list. Make sure those motions come through as usual, but
                # also add an alias so we can override the motions with an edited
                # version of the motion.
                if block["AnimID"] in STATIC_VFS:
                    new_block = copy.deepcopy(block)
                    new_block["AnimID"] = tracker.get_alias_uuid(anim_id)
                    message["AnimationList"].append(new_block)
                else:
                    block["AnimID"] = tracker.get_alias_uuid(anim_id)
        elif message.name == "AgentAnimation":
            # Make sure to remove any alias IDs from our outbound anim requests
            for block in message["AnimationList"]:
                orig_id = tracker.get_orig_uuid(block["AnimID"])
                if orig_id:
                    block["AnimID"] = orig_id

    def handle_http_request(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        if not flow.cap_data.asset_server_cap:
            return

        anim_id = flow.request.query.get("animatn_id")
        if not anim_id:
            return

        orig_anim_id = self.horror_anim_tracker.get_orig_uuid(UUID(anim_id))
        if not orig_anim_id:
            return

        flow.request.query["animatn_id"] = str(orig_anim_id)

        flow.can_stream = False
        flow.metadata["horror_anim"] = True

        if orig_anim_id in STATIC_VFS:
            # These animations are only in the static VFS and won't be served
            # by the asset server. Read the anim out of the static VFS and
            # send the response back immediately
            block = STATIC_VFS[orig_anim_id]
            anim_data = STATIC_VFS.read_block(block)
            flow.response = mitmproxy.http.Response.make(
                200,
                _mutate_anim_bytes(anim_data),
                {
                    "Content-Type": "binary/octet-stream",
                    "Connection": "close",
                }
            )
            return True

        # Partial requests for an anim wouldn't make any sense
        flow.request.headers.pop("Range", None)

    def handle_http_response(self, session_manager: SessionManager, flow: HippoHTTPFlow):
        if not flow.metadata.get("horror_anim"):
            return

        if flow.response.status_code not in (200, 206):
            return

        flow.response.content = _mutate_anim_bytes(flow.response.content)
        # Not a range anymore, update the headers and status.
        flow.response.headers.pop("Content-Range", None)
        flow.response.status_code = 200

        return True


addons = [HorrorAnimatorAddon()]



from __future__ import annotations

import itertools
import random
import struct
import unittest
from typing import *

import lazy_object_proxy

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.message_logger import FilteringMessageLogger, LLUDPMessageLogEntry
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session
from hippolyzer.lib.proxy.test_utils import BaseProxyTest


UNKNOWN_PACKET = b'\x00\x00\x00\x00E\x00\xff\xf0\x00\xff\xff\xff\xff\x00'


class MockAddon(BaseAddon):
    def __init__(self):
        self.events = []

    def handle_session_init(self, session: Session):
        self.events.append(("session_init", session.id))

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        self.events.append(("lludp", session.id, region.circuit_addr, message.name))
        if message.name == "UndoLand":
            # Simulate a message being taken out of the regular proxying flow
            message.take()
            return True

    def handle_object_updated(self, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        self.events.append(("object_update", session.id, region.circuit_addr, obj.LocalID, updated_props))


class SimpleMessageLogger(FilteringMessageLogger):
    @property
    def entries(self):
        return self._filtered_entries


class LLUDPIntegrationTests(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.addon = MockAddon()
        self.deserializer = UDPMessageDeserializer()
        AddonManager.init([], self.session_manager, [self.addon])

    def _make_objectupdate_compressed(self, localid: Optional[int] = None, handle: Optional[int] = 123):
        if localid is None:
            localid = random.getrandbits(32)

        return b'\x00\x00\x00\x0c\xba\x00\r' + struct.pack("<Q", handle) + b'\xff\xff\x03\xd0\x04\x00\x10' \
               b'\xe6\x00\x12\x12\x10\xbf\x16XB~\x8f\xb4\xfb\x00\x1a\xcd\x9b\xe5' + struct.pack("<I", localid) + \
               b'\t\x00\xcdG\x00\x00\x03\x00\x00\x00\x1cB\x00\x00\x1cB\xcd\xcc\xcc=\xedG,' \
               b'B\x9e\xb1\x9eBff\xa0A\x00\x00\x00\x00\x00\x00\x00\x00[' \
               b'\x8b\xf8\xbe\xc0\x00\x00\x00k\x9b\xc4\xfe3\nOa\xbb\xe2\xe4\xb2C\xac7\xbd\x00\x00\x00\x00\x00' \
               b'\x00\x00\x00\x00\x00\xa2=\x010\x00\x11\x00\x00\x00\x89UgG$\xcbC\xed\x92\x0bG\xca\xed' \
               b'\x15F_@ \x00\x00\x00\x00d\x96\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
               b'\x00?\x00\x00\x00\x1c\x9fJoI\x8dH\xa0\x9d\xc4&\'\'\x19=g\x00\x00\x00\x003\x00ff\x86\xbf' \
               b'\x00ff\x86?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89UgG$\xcbC' \
               b'\xed\x92\x0bG\xca\xed\x15F_\x10\x00\x00\x003\x00\x01\x01\x00\x00\x00\x00\xdb\x0f\xc9@\xa6' \
               b'\x9b\xc4=\xd0\x04\x00\x10\xe6\x00\xc2\xa62\xe2\x9b\xd7L\xc4\xbb\xd6\x1fKC\xa6\xdf\x8d\\\x04\x00' \
               b'\x00\t\x00\xd3G\x00\x00\x03\x00\x00\x00\x1cB\x00\x00\x1cB\xcd\xcc\xcc=\t\x08\x9cA\xf2\x03' \
               b'\xa5Bff\xa0A\x00\x00\x00\x00\x00\x00\x00\x00[\x8b\xf8' \
               b'\xbe\xc0\x00\x00\x00\x0b\x1b\xa0\xd1\x97=C\xcd\xae\x19\xfd\xc9\xbb\x88\x05\xc3\x00\x00\x00\x00\x00' \
               b'\x00\x00\x00\x00\x00\xa2=\x010\x00\x11\x00\x00\x00\x89UgG$\xcbC\xed\x92\x0bG\xca\xed' \
               b'\x15F_@ \x00\x00\x00\x00d\x96\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
               b'\x00?\x00\x00\x00\xbd\x8b\xd7h{\xdbM\xbc\x8c3X\xa6\xa6\x0c\x94\xd7\x00\x00\x00\x003\x00ff\x86\xbf' \
               b'\x00ff\x86?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89UgG$\xcbC' \
               b'\xed\x92\x0bG\xca\xed\x15F_\x10\x00\x00\x003\x00\x01\x01\x00\x00\x00\x00\xdb\x0f\xc9@\xa6' \
               b'\x9b\xc4=\xd0\x04\x00\x10\xe6\x00\xd1e\xac\xff,NBK\x91d\xbb\x15\\\x0b\xc3\x9c\xe2\x05\x00\x00' \
               b'\t\x00\xbbG\x00\x00\x03\x00\x00\x00\x1cB\x00\x00\x1cB\xcd\xcc\xcc=\x0f5\x97AY\x98ZBff' \
               b'\xa0A\x00\x00\x00\x00\x00\x00\x00\x00\xe6Y0\xbf\xc0\x00\x00\x00\x89UgG$\xcbC\xed\x92\x0bG' \
               b'\xca\xed\x15F_\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa2=\x010\x00\x11\x00\x00\x00' \
               b'#\xce\xf8\xf4\x0cJD.\xb7"\x96\x1cK\xd9\x01\x1b@ ' \
               b'\x00\x00\x00\x00d\x96\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
               b'?\x00\x00\x003\xe1\xa1\xcf<\xbdD\xc4\xa0\xe6b\xe9\xbf=\xa2@\x00\x00\x00\x003\x00ff\x86\xbf' \
               b'\x00ff\x86?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89UgG$\xcbC' \
               b'\xed\x92\x0bG\xca\xed\x15F_\x10\x00\x00\x003\x00\x01\x01\x00\x00\x00\x00\xdb\x0f\xc9@\xa6' \
               b'\x9b\xc4='

    async def test_session_claiming(self):
        # Need a UseCircuitCode to claim a pending session
        msg = Message(
            "UseCircuitCode",
            Block("CircuitCode", Code=self.circuit_code, SessionID=self.session.id,
                  ID=self.session.agent_id),
            packet_id=1,
        )
        datagram = self._msg_to_datagram(msg, self.client_addr, self.region_addr,
                                         socks_header=True)
        self.protocol.datagram_received(datagram, self.client_addr)
        await self._wait_drained()
        self.assertFalse(self.session.pending)
        # session should now be tied to the protocol
        self.assertEqual(self.session, self.protocol.session)
        # Packet got forwarded through
        self.assertEqual(len(self.transport.packets), 1)

    async def test_bad_session_unclaimed(self):
        # Need a UseCircuitCode to claim a pending session
        msg = Message(
            "UseCircuitCode",
            Block("CircuitCode", Code=self.circuit_code, SessionID=UUID.random(),
                  ID=self.session.agent_id),
            packet_id=1,
        )
        datagram = self._msg_to_datagram(msg, self.client_addr, self.region_addr,
                                         socks_header=True)
        self.protocol.datagram_received(datagram, source_addr=self.client_addr)
        await self._wait_drained()
        # Packet got dropped completely
        self.assertTrue(self.session.pending)
        self.assertIsNone(self.protocol.session)
        self.assertEqual(len(self.transport.packets), 0)

    async def test_bad_circuit_not_sent(self):
        # Need a UseCircuitCode to claim a pending session
        msg = Message(
            "UseCircuitCode",
            Block("CircuitCode", Code=self.circuit_code, SessionID=self.session.id,
                  ID=self.session.agent_id),
            packet_id=1,
        )
        datagram = self._msg_to_datagram(msg, self.client_addr, (self.region_addr[0], 9),
                                         socks_header=True)
        self.protocol.datagram_received(datagram, source_addr=self.client_addr)
        await self._wait_drained()
        # The session claim will still work
        self.assertFalse(self.session.pending)
        self.assertEqual(self.session, self.protocol.session)
        # Packet should not get sent, though and should not be handled
        self.assertEqual(len(self.transport.packets), 0)

    @unittest.skip("expensive")
    async def test_benchmark(self):
        num_times = 10_000
        self._setup_default_circuit()
        for _ in range(num_times):
            self.protocol.datagram_received(self._make_objectupdate_compressed(), self.region_addr)
        await self._wait_drained()
        self.assertEqual(len(self.transport.packets), num_times)

    async def test_forwarded_with_header(self):
        self._setup_default_circuit()
        obj_update = self._make_objectupdate_compressed()
        self.protocol.datagram_received(obj_update, self.region_addr)
        await self._wait_drained()
        packets = self.transport.packets
        self.assertEqual(len(packets), 1)
        data, dst_addr = packets[0]
        # Was being sent towards the client
        self.assertEqual(dst_addr, self.client_addr)
        # The data should not have changed since we haven't injected
        # or dropped anything
        self.assertEqual(obj_update, data)

    async def test_addon_hooks(self):
        self._setup_default_circuit()
        obj_update = self._make_objectupdate_compressed()
        self.protocol.datagram_received(obj_update, self.region_addr)
        await self._wait_drained()
        expected_lludp_event = ("lludp", self.session.id, self.region_addr, "ObjectUpdateCompressed")
        self.assertTrue(any(x == expected_lludp_event for x in self.addon.events))

    async def test_object_added_with_tes(self):
        self._setup_default_circuit()
        obj_update = self._make_objectupdate_compressed(1234)
        self.protocol.datagram_received(obj_update, self.region_addr)
        await self._wait_drained()
        obj = self.session.regions[0].objects.lookup_localid(1234)
        self.assertIsInstance(obj.TextureEntry, lazy_object_proxy.Proxy)
        self.assertEqual(obj.TextureEntry.Textures[None], UUID("1c9f4a6f-498d-48a0-9dc4-262727193d67"))
        self.assertEqual(len(self.session.regions[0].objects), 3)

    async def test_object_updated_changed_property_list(self):
        self._setup_default_circuit()
        # One creating update and one no-op update
        obj_update = self._make_objectupdate_compressed(1234)
        self.protocol.datagram_received(obj_update, self.region_addr)
        obj_update = self._make_objectupdate_compressed(1234)
        self.protocol.datagram_received(obj_update, self.region_addr)
        await self._wait_drained()
        self.assertEqual(3, len(self.session.regions[0].objects))
        object_events = [e for e in self.addon.events if e[0] == "object_update"]
        # 3 objects in example packet and we sent it twice
        self.assertEqual(6, len(object_events))
        # Only TextureEntry should be marked updated since it's a proxy object
        self.assertEqual(object_events[-1][-1], {"TextureEntry"})

    async def test_message_logger(self):
        message_logger = SimpleMessageLogger()
        self.session_manager.message_logger = message_logger
        self._setup_default_circuit()
        obj_update = self._make_objectupdate_compressed(1234)
        self.protocol.datagram_received(obj_update, self.region_addr)
        await self._wait_drained()
        entries = message_logger.entries
        self.assertEqual(1, len(entries))
        self.assertEqual("ObjectUpdateCompressed", entries[0].name)

    async def test_filtering_logged_messages(self):
        message_logger = SimpleMessageLogger()
        self.session_manager.message_logger = message_logger
        self._setup_default_circuit()
        obj_update = self._make_objectupdate_compressed(1234)
        self.protocol.datagram_received(obj_update, self.region_addr)
        msg = self.serializer.serialize(Message(
            "UndoLand",
            Block("AgentData", AgentID=UUID(), SessionID=UUID()),
            direction=Direction.IN,
        ))
        self.protocol.datagram_received(msg, self.region_addr)
        await self._wait_drained()
        message_logger.set_filter("ObjectUpdateCompressed")
        entries = message_logger.entries
        self.assertEqual(1, len(entries))
        self.assertEqual("ObjectUpdateCompressed", entries[0].name)

    async def test_logging_taken_message(self):
        message_logger = SimpleMessageLogger()
        self.session_manager.message_logger = message_logger
        self._setup_default_circuit()
        msg = self.serializer.serialize(Message(
            "UndoLand",
            Block("AgentData", AgentID=UUID(), SessionID=UUID()),
            direction=Direction.IN,
        ))
        self.protocol.datagram_received(msg, self.region_addr)
        await self._wait_drained()
        entries = message_logger.entries
        self.assertEqual(len(entries), 1)
        entry: LLUDPMessageLogEntry = entries[0]  # type: ignore
        self.assertEqual(entry.name, "UndoLand")
        self.assertEqual(entry.message.dropped, True)

    async def test_logging_unknown_message(self):
        message_logger = SimpleMessageLogger()
        self.session_manager.message_logger = message_logger
        self._setup_default_circuit()
        self.protocol.datagram_received(UNKNOWN_PACKET, self.region_addr)
        await self._wait_drained()
        entries = message_logger.entries
        self.assertEqual(len(entries), 1)
        entry: LLUDPMessageLogEntry = entries[0]  # type: ignore
        # Freezing shouldn't affect this
        entry.freeze()
        self.assertEqual(entry.name, "UnknownMessage:240")
        self.assertEqual(entry.message.dropped, False)
        self.assertEqual(entry.message.unknown_message, True)

    async def test_session_message_handler(self):
        self._setup_default_circuit()
        obj_update = self._make_objectupdate_compressed(1234)
        fut = self.session.message_handler.wait_for(('ObjectUpdateCompressed',))
        self.protocol.datagram_received(obj_update, self.region_addr)
        self.assertEqual("ObjectUpdateCompressed", (await fut).name)

    def test_roundtrip_objectupdatecompressed(self):
        msg_bytes = self._make_objectupdate_compressed()
        message: Message = self.deserializer.deserialize(msg_bytes)
        for block in itertools.chain(*message.blocks.values()):
            for var_name in block.vars.keys():
                orig_val = block[var_name]
                try:
                    serializer = block.get_serializer(var_name)
                except KeyError:
                    # Don't have a serializer, onto the next field
                    continue
                deser = serializer.deserialize(block, orig_val)
                new_val = serializer.serialize(block, deser)
                if orig_val != new_val:
                    raise AssertionError(f"{block.name}.{var_name} didn't reserialize correctly,"
                                         f"{orig_val!r} != {new_val!r}")
        new_msg_bytes = self.serializer.serialize(message)
        self.assertEqual(new_msg_bytes, msg_bytes)



import copy
import datetime as dt
import unittest

from hippolyzer.lib.base.datatypes import *
from hippolyzer.lib.base.inventory import InventoryModel, SaleType, InventoryItem
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.wearables import Wearable, VISUAL_PARAMS

SIMPLE_INV = """\tinv_object\t0
\t{
\t\tobj_id\tf4d91477-def1-487a-b4f3-6fa201c17376
\t\tparent_id\t00000000-0000-0000-0000-000000000000
\t\ttype\tcategory
\t\tname\tContents|
\t\tmetadata\t<llsd><undef /></llsd>
|
\t}
\tinv_item\t0
\t{
\t\titem_id\tdd163122-946b-44df-99f6-a6030e2b9597
\t\tparent_id\tf4d91477-def1-487a-b4f3-6fa201c17376
\tpermissions 0
\t{
\t\tbase_mask\t7fffffff
\t\towner_mask\t7fffffff
\t\tgroup_mask\t00000000
\t\teveryone_mask\t00000000
\t\tnext_owner_mask\t0008e000
\t\tcreator_id\ta2e76fcd-9360-4f6d-a924-000000000003
\t\towner_id\ta2e76fcd-9360-4f6d-a924-000000000003
\t\tlast_owner_id\ta2e76fcd-9360-4f6d-a924-000000000003
\t\tgroup_id\t00000000-0000-0000-0000-000000000000
\t}
\t\tasset_id\t00000000-0000-0000-0000-000000000000
\t\ttype\tlsltext
\t\tinv_type\tscript
\t\tflags\t00000000
\tsale_info\t0
\t{
\t\tsale_type\tnot
\t\tsale_price\t10
\t}
\t\tname\tNew Script|
\t\tdesc\t2020-04-20 04:20:39 lsl2 script|
\t\tmetadata\t<llsd><map><key>experience</key><uuid>a2e76fcd-9360-4f6d-a924-000000000003</uuid></map></llsd>
|
\t\tcreation_date\t1587367239
\t}
"""

SIMPLE_INV_PARSED = [
    {
        'name': 'Contents',
        'obj_id': UUID('f4d91477-def1-487a-b4f3-6fa201c17376'),
        'parent_id': UUID('00000000-0000-0000-0000-000000000000'),
        'type': 'category'
    },
    {
        'asset_id': UUID('00000000-0000-0000-0000-000000000000'),
        'created_at': 1587367239,
        'desc': '2020-04-20 04:20:39 lsl2 script',
        'flags': b'\x00\x00\x00\x00',
        'inv_type': 'script',
        'item_id': UUID('dd163122-946b-44df-99f6-a6030e2b9597'),
        'name': 'New Script',
        'metadata': {"experience": UUID("a2e76fcd-9360-4f6d-a924-000000000003")},
        'parent_id': UUID('f4d91477-def1-487a-b4f3-6fa201c17376'),
        'permissions': {
            'base_mask': 2147483647,
            'creator_id': UUID('a2e76fcd-9360-4f6d-a924-000000000003'),
            'everyone_mask': 0,
            'group_id': UUID('00000000-0000-0000-0000-000000000000'),
            'group_mask': 0,
            'last_owner_id': UUID('a2e76fcd-9360-4f6d-a924-000000000003'),
            'next_owner_mask': 581632,
            'owner_id': UUID('a2e76fcd-9360-4f6d-a924-000000000003'),
            'owner_mask': 2147483647,
        },
        'sale_info': {
            'sale_price': 10,
            'sale_type': 'not'
        },
        'type': 'lsltext'
    }
]

INV_CATEGORY = """\tinv_category\t0
\t{
\t\tcat_id\tf4d91477-def1-487a-b4f3-6fa201c17376
\t\tparent_id\t00000000-0000-0000-0000-000000000000
\t\ttype\tlsltext
\t\tpref_type\tlsltext
\t\tname\tScripts|
\t\towner_id\ta2e76fcd-9360-4f6d-a924-000000000003
\t}
"""


class TestLegacyInv(unittest.TestCase):
    def setUp(self) -> None:
        self.model = InventoryModel.from_str(SIMPLE_INV)

    def test_parse(self):
        self.assertTrue(UUID('f4d91477-def1-487a-b4f3-6fa201c17376') in self.model.nodes)
        self.assertIsNotNone(self.model.root)

    def test_parse_category(self):
        model = InventoryModel.from_str(INV_CATEGORY)
        self.assertEqual(UUID('f4d91477-def1-487a-b4f3-6fa201c17376'), model.root.node_id)

    def test_serialize(self):
        new_model = InventoryModel.from_str(self.model.to_str())
        self.assertEqual(self.model, new_model)

    def test_serialize_category(self):
        model = InventoryModel.from_str(INV_CATEGORY)
        new_model = InventoryModel.from_str(model.to_str())
        self.assertEqual(model, new_model)

    def test_category_legacy_serialization(self):
        self.assertEqual(INV_CATEGORY, InventoryModel.from_str(INV_CATEGORY).to_str())

    def test_item_access(self):
        item = self.model.nodes[UUID('dd163122-946b-44df-99f6-a6030e2b9597')]
        self.assertEqual(item.name, "New Script")
        self.assertEqual(item.sale_info.sale_type, SaleType.NOT)
        self.assertDictEqual(item.metadata, {"experience": UUID("a2e76fcd-9360-4f6d-a924-000000000003")})
        self.assertEqual(item.model, self.model)

    def test_access_children(self):
        root = self.model.root
        item = tuple(self.model.ordered_nodes)[1]
        self.assertEqual((item,), root.children)

    def test_access_parent(self):
        root = self.model.root
        item = tuple(self.model.ordered_nodes)[1]
        self.assertEqual(root, item.parent)
        self.assertEqual(None, root.parent)

    def test_unlink(self):
        self.assertEqual(1, len(self.model.root.children))
        item = tuple(self.model.ordered_nodes)[1]
        self.assertEqual([item], item.unlink())
        self.assertEqual(0, len(self.model.root.children))
        self.assertEqual(None, item.model)

    def test_relink(self):
        item = tuple(self.model.ordered_nodes)[1]
        for unlinked in item.unlink():
            self.model.add(unlinked)
        self.assertEqual(self.model, item.model)
        self.assertEqual(1, len(self.model.root.children))

    def test_eq_excludes_model(self):
        item = tuple(self.model.ordered_nodes)[1]
        item_copy = copy.copy(item)
        item_copy.model = None
        self.assertEqual(item, item_copy)

    def test_llsd_serialization(self):
        self.assertEqual(self.model.to_llsd(), SIMPLE_INV_PARSED)

    def test_llsd_date_parsing(self):
        model = InventoryModel.from_llsd(SIMPLE_INV_PARSED)
        item: InventoryItem = model.nodes.get(UUID("dd163122-946b-44df-99f6-a6030e2b9597"))  # type: ignore
        self.assertEqual(item.creation_date, dt.datetime(2020, 4, 20, 7, 20, 39, tzinfo=dt.timezone.utc))

    def test_llsd_serialization_ais(self):
        model = InventoryModel.from_str(INV_CATEGORY)
        self.assertEqual(
            [
                {
                    'agent_id': UUID('a2e76fcd-9360-4f6d-a924-000000000003'),
                    'category_id': UUID('f4d91477-def1-487a-b4f3-6fa201c17376'),
                    'name': 'Scripts',
                    'parent_id': UUID('00000000-0000-0000-0000-000000000000'),
                    'type_default': 10,
                    'version': -1
                }
            ],
            model.to_llsd("ais")
        )

    def test_llsd_legacy_equality(self):
        new_model = InventoryModel.from_llsd(self.model.to_llsd())
        self.assertEqual(self.model, new_model)
        new_model.root.name = "foo"
        self.assertNotEqual(self.model, new_model)

    def test_legacy_serialization(self):
        self.assertEqual(SIMPLE_INV, self.model.to_str())

    def test_difference_added(self):
        new_model = InventoryModel.from_llsd(self.model.to_llsd())
        diff = self.model.get_differences(new_model)
        self.assertEqual([], diff.changed)
        self.assertEqual([], diff.removed)

        new_model.root.name = "foo"
        diff = self.model.get_differences(new_model)
        self.assertEqual([new_model.root], diff.changed)
        self.assertEqual([], diff.removed)

        item = new_model.root.children[0]
        item.unlink()
        diff = self.model.get_differences(new_model)
        self.assertEqual([new_model.root], diff.changed)
        self.assertEqual([item], diff.removed)

        new_item = copy.copy(item)
        new_item.node_id = UUID.random()
        new_model.add(new_item)
        diff = self.model.get_differences(new_model)
        self.assertEqual([new_model.root, new_item], diff.changed)
        self.assertEqual([item], diff.removed)


GIRL_NEXT_DOOR_SHAPE = """LLWearable version 22
Girl Next Door - C2 - med - Adam n Eve

\tpermissions 0
\t{
\t\tbase_mask\t0008e000
\t\towner_mask\t0008e000
\t\tgroup_mask\t00000000
\t\teveryone_mask\t00000000
\t\tnext_owner_mask\t0008e000
\t\tcreator_id\tdf110c10-72e6-40d6-8916-14b4b0366b18
\t\towner_id\t285a8af7-7a1d-4608-8fb5-200b816e7ee1
\t\tlast_owner_id\t285a8af7-7a1d-4608-8fb5-200b816e7ee1
\t\tgroup_id\t00000000-0000-0000-0000-000000000000
\t}
\tsale_info\t0
\t{
\t\tsale_type\tnot
\t\tsale_price\t10
\t}
type 0
parameters 82
1 -.21
2 -.4
4 -.08
5 .04
6 -.04
7 .02
8 -.42
10 -.15
11 .16
12 -.5
13 0
14 1
15 -.16
17 .17
18 -.42
19 -.78
20 -.16
21 .01
22 0
23 -.44
24 .32
25 .35
27 -.55
33 -.11
34 .22
35 -.1
36 -1.06
37 -.86
38 -.42
80 0
105 .59
155 -.24
157 0
185 -.68
193 .55
196 -.98
505 .39
506 -.68
507 -.77
515 -1
517 -.11
518 .6
629 .5
637 .24
646 -.26
647 -.05
649 .36
650 .25
652 .4
653 .29
656 0
659 .76
662 .5
663 0
664 0
665 .08
675 -.14
676 -.58
678 .5
682 .5
683 -.17
684 .28
685 0
690 .42
692 .26
693 -.12
753 .51
756 -.2
758 .21
759 .5
760 .85
764 .32
765 .18
769 .37
773 .53
795 .22
796 0
799 .5
841 -.12
842 0
879 0
880 -.37
textures 0
"""

# TODO: Move appearance-related stuff elsewhere.

GIRL_NEXT_DOOR_APPEARANCE_MSG = Message(
    'AvatarAppearance',
    Block('Sender', ID=UUID(int=1), IsTrial=0),
    # We don't care about the value of this.
    Block('ObjectData', TextureEntry=b""),
    Block('VisualParam', ParamValue=9),
    Block('VisualParam', ParamValue=30),
    Block('VisualParam', ParamValue=71),
    Block('VisualParam', ParamValue=32),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=132),
    Block('VisualParam', ParamValue=10),
    Block('VisualParam', ParamValue=76),
    Block('VisualParam', ParamValue=84),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=43),
    Block('VisualParam', ParamValue=83),
    Block('VisualParam', ParamValue=113),
    Block('VisualParam', ParamValue=68),
    Block('VisualParam', ParamValue=73),
    Block('VisualParam', ParamValue=43),
    Block('VisualParam', ParamValue=35),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=7),
    Block('VisualParam', ParamValue=132),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=76),
    Block('VisualParam', ParamValue=91),
    Block('VisualParam', ParamValue=129),
    Block('VisualParam', ParamValue=106),
    Block('VisualParam', ParamValue=76),
    Block('VisualParam', ParamValue=58),
    Block('VisualParam', ParamValue=99),
    Block('VisualParam', ParamValue=73),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=203),
    Block('VisualParam', ParamValue=48),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=150),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=114),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=76),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=40),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=140),
    Block('VisualParam', ParamValue=86),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=99),
    Block('VisualParam', ParamValue=84),
    Block('VisualParam', ParamValue=53),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=66),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=100),
    Block('VisualParam', ParamValue=216),
    Block('VisualParam', ParamValue=214),
    Block('VisualParam', ParamValue=204),
    Block('VisualParam', ParamValue=204),
    Block('VisualParam', ParamValue=204),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=89),
    Block('VisualParam', ParamValue=109),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=61),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=115),
    Block('VisualParam', ParamValue=76),
    Block('VisualParam', ParamValue=91),
    Block('VisualParam', ParamValue=158),
    Block('VisualParam', ParamValue=102),
    Block('VisualParam', ParamValue=109),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=193),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=132),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=68),
    Block('VisualParam', ParamValue=35),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=97),
    Block('VisualParam', ParamValue=92),
    Block('VisualParam', ParamValue=79),
    Block('VisualParam', ParamValue=107),
    Block('VisualParam', ParamValue=160),
    Block('VisualParam', ParamValue=112),
    Block('VisualParam', ParamValue=63),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=159),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=73),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=102),
    Block('VisualParam', ParamValue=158),
    Block('VisualParam', ParamValue=145),
    Block('VisualParam', ParamValue=153),
    Block('VisualParam', ParamValue=163),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=122),
    Block('VisualParam', ParamValue=43),
    Block('VisualParam', ParamValue=94),
    Block('VisualParam', ParamValue=135),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=214),
    Block('VisualParam', ParamValue=204),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=56),
    Block('VisualParam', ParamValue=30),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=204),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=112),
    Block('VisualParam', ParamValue=127),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=100),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=84),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=94),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=255),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=23),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=23),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=23),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=23),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=23),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=0),
    Block('VisualParam', ParamValue=25),
    Block('VisualParam', ParamValue=23),
    Block('VisualParam', ParamValue=51),
    Block('VisualParam', ParamValue=1),
    Block('VisualParam', ParamValue=127),
    Block('AppearanceData', AppearanceVersion=1, CofVersion=100, Flags=0),
    Block('AppearanceHover', HoverHeight=Vector3(0.0, 0.0, 0.0))
)


class TestWearable(unittest.TestCase):
    def test_parse(self):
        wearable = Wearable.from_str(GIRL_NEXT_DOOR_SHAPE)
        self.assertEqual(wearable.name, "Girl Next Door - C2 - med - Adam n Eve")
        self.assertEqual(wearable.parameters[841], -.12)

    def test_serialize(self):
        wearable = Wearable.from_str(GIRL_NEXT_DOOR_SHAPE)
        new_wearable = Wearable.from_str(wearable.to_str())
        self.assertEqual(wearable, new_wearable)

    def test_visual_params(self):
        param = VISUAL_PARAMS.by_name("Eyelid_Inner_Corner_Up")
        self.assertEqual(param.value_max, 1.2)

    def test_message_equivalent(self):
        wearable = Wearable.from_str(GIRL_NEXT_DOOR_SHAPE)
        parsed = VISUAL_PARAMS.parse_appearance_message(GIRL_NEXT_DOOR_APPEARANCE_MSG)

        for i, (param_id, param_val) in enumerate(parsed.items()):
            param = VISUAL_PARAMS.by_id(param_id)
            if param.wearable != "shape":
                continue
            # A parameter may legitimately be missing from the shape depending on its age,
            # just assume it's the default value.
            expected_val = wearable.parameters.get(param_id, param.value_default)
            # This seems like quite a large delta. Maybe we should be using different quantization here.
            self.assertAlmostEqual(expected_val, param_val, delta=0.015)



import asyncio
import datetime as dt
from typing import *

from hippolyzer.lib.proxy.addon_utils import send_chat, BaseAddon, SessionProperty
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class TaskExampleAddon(BaseAddon):
    chat_loop_task: Optional[asyncio.Task] = SessionProperty(None)
    chat_loop_count: Optional[int] = SessionProperty(None)
    chat_loop_started: Optional[dt.datetime] = SessionProperty(None)

    @handle_command()
    async def start_chat_task(self, session: Session, _region: ProxiedRegion):
        """Start a task that sends chat in a loop, demonstrating task scheduling"""
        # Already doing a chat loop
        if self.chat_loop_task and not self.chat_loop_task.done():
            return
        self.chat_loop_started = dt.datetime.now()
        self.chat_loop_count = 0
        # Don't need to clean this up on session shutdown because _schedule_task()
        # binds tasks to session lifetime by default
        self.chat_loop_task = self._schedule_task(self._chat_loop(session))

    @handle_command()
    async def stop_chat_task(self, _session: Session, _region: ProxiedRegion):
        """Stop the chat task if one was active"""
        if self.chat_loop_task and not self.chat_loop_task.done():
            self.chat_loop_task.cancel()
        self.chat_loop_task = None

    async def _chat_loop(self, session: Session, sleep_time=5.0):
        while True:
            region = session.main_region
            if not region:
                await asyncio.sleep(sleep_time)
                continue
            send_chat(
                f"Loop {self.chat_loop_count}, started "
                f"{dt.datetime.now() - self.chat_loop_started} ago",
                session=session
            )
            self.chat_loop_count += 1
            await asyncio.sleep(sleep_time)


addons = [TaskExampleAddon()]



"""
Loading task inventory doesn't actually need to be slow.

By using a cap instead of the slow xfer path and sending the LLSD inventory
model we get 15x speedups even when mocking things behind the scenes by using
a hacked up version of xfer. See turbo_object_inventory.py
"""

import asyncio

import asgiref.wsgi
from typing import *

from flask import Flask, Response, request

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.inventory import InventoryModel, InventoryObject
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.templates import XferFilePath, AssetType
from hippolyzer.lib.proxy import addon_ctx
from hippolyzer.lib.proxy.webapp_cap_addon import WebAppCapAddon

app = Flask("GetTaskInventoryCapApp")


@app.route('/', methods=["GET"])
async def get_task_inventory():
    # Should always have the current region, the cap handler is bound to one.
    # Just need to pull it from the `addon_ctx` module's global.
    region = addon_ctx.region.get()
    session = addon_ctx.session.get()
    obj_id = UUID(request.args["task_id"])
    obj = region.objects.lookup_fullid(obj_id)
    if not obj:
        return Response(f"Couldn't find {obj_id}", status=404, mimetype="text/plain")
    request_msg = Message(
        'RequestTaskInventory',
        Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
        Block('InventoryData', LocalID=obj.LocalID),
    )
    # Keep around a dict of chunks we saw previously in case we have to restart
    # an Xfer due to missing chunks. We don't expect chunks to change across Xfers
    # so this can be used to recover from dropped SendXferPackets in subsequent attempts
    existing_chunks: Dict[int, bytes] = {}
    for _ in range(3):
        # Any previous requests will have triggered a delete of the inventory file
        # by marking it complete on the server-side. Re-send our RequestTaskInventory
        # To make sure there's a fresh copy.
        region.circuit.send(request_msg.take())
        inv_message = await region.message_handler.wait_for(
            ('ReplyTaskInventory',),
            predicate=lambda x: x["InventoryData"]["TaskID"] == obj.FullID,
            timeout=5.0,
        )
        # No task inventory, send the reply as-is
        file_name = inv_message["InventoryData"]["Filename"]
        if not file_name:
            # The "Contents" folder always has to be there, if we don't put it here
            # then the viewer will have to lie about it being there itself.
            return Response(
                llsd.format_xml({
                    "inventory": [
                        InventoryObject(
                            name="Contents",
                            parent_id=UUID.ZERO,
                            type=AssetType.CATEGORY,
                            obj_id=obj_id
                        ).to_llsd()
                    ],
                    "inv_serial": inv_message["InventoryData"]["Serial"],
                }),
                headers={"Content-Type": "application/llsd+xml"},
                status=200,
            )

        last_serial = request.args.get("last_serial", None)
        if last_serial:
            last_serial = int(last_serial)
        if inv_message["InventoryData"]["Serial"] == last_serial:
            # Nothing has changed since the version of the inventory they say they have, say so.
            return Response("", status=304)

        xfer = region.xfer_manager.request(
            file_name=file_name,
            file_path=XferFilePath.CACHE,
            turbo=True,
        )
        xfer.chunks.update(existing_chunks)
        try:
            await xfer
        except asyncio.TimeoutError:
            # We likely failed the request due to missing chunks, store
            # the chunks that we _did_ get for the next attempt.
            existing_chunks.update(xfer.chunks)
            continue

        inv_model = InventoryModel.from_str(xfer.reassemble_chunks().decode("utf8"))

        return Response(
            llsd.format_xml({
                "inventory": inv_model.to_llsd(),
                "inv_serial": inv_message["InventoryData"]["Serial"],
            }),
            headers={"Content-Type": "application/llsd+xml"},
        )
    raise asyncio.TimeoutError("Failed to get inventory after 3 tries")


class GetTaskInventoryCapExampleAddon(WebAppCapAddon):
    # A cap URL with this name will be tied to each region when
    # the sim is first connected to. The URL will be returned to the
    # viewer in the Seed if the viewer requests it by name.
    CAP_NAME = "GetTaskInventoryExample"
    # Any asgi app should be fine.
    APP = asgiref.wsgi.WsgiToAsgi(app)


addons = [GetTaskInventoryCapExampleAddon()]



"""
Base classes for common session-related state shared between clients and proxies
"""
from __future__ import annotations

import abc
import logging
import weakref
from typing import *

import multidict

from hippolyzer.lib.base.datatypes import UUID, Vector3
from hippolyzer.lib.base.message.circuit import ConnectionHolder, Circuit
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.network.caps_client import CapsClient
from hippolyzer.lib.base.network.transport import ADDR_TUPLE
from hippolyzer.lib.base.objects import handle_to_global_pos
from hippolyzer.lib.base.xfer_manager import XferManager
from hippolyzer.lib.client.inventory_manager import InventoryManager

from hippolyzer.lib.client.object_manager import ClientObjectManager, ClientWorldObjectManager


class BaseClientRegion(ConnectionHolder, abc.ABC):
    """Represents a client's view of a remote region"""
    handle: Optional[int]
    # Actually a weakref
    session: Callable[[], BaseClientSession]
    objects: ClientObjectManager
    xfer_manager: XferManager
    caps_client: CapsClient
    cap_urls: multidict.MultiDict[str]
    circuit_addr: ADDR_TUPLE
    circuit: Optional[Circuit]
    _name: Optional[str]

    def __init__(self):
        self._name = None
        self.circuit = None

    @abc.abstractmethod
    def update_caps(self, caps: Mapping[str, str]) -> None:
        pass

    @property
    def name(self):
        if self._name:
            return self._name
        return "Pending %r" % (self.circuit_addr,)

    @name.setter
    def name(self, val):
        self._name = val

    @property
    def global_pos(self) -> Vector3:
        if self.handle is None:
            raise ValueError("Can't determine global region position without handle")
        return handle_to_global_pos(self.handle)

    @property
    def is_alive(self):
        if not self.circuit:
            return False
        return self.circuit.is_alive

    def mark_dead(self):
        logging.info("Marking %r dead" % self)
        if self.circuit:
            self.circuit.is_alive = False
        self.objects.clear()

    def __repr__(self):
        return "<%s %s (%r)>" % (self.__class__.__name__, self.name, self.handle)


class BaseClientSessionManager:
    pass


class BaseClientSession(abc.ABC):
    """Represents a client's view of a remote session"""
    id: UUID
    agent_id: UUID
    secure_session_id: UUID
    active_group: UUID
    groups: Set[UUID]
    message_handler: MessageHandler[Message, str]
    regions: MutableSequence[BaseClientRegion]
    region_by_handle: Callable[[int], Optional[BaseClientRegion]]
    region_by_circuit_addr: Callable[[ADDR_TUPLE], Optional[BaseClientRegion]]
    objects: ClientWorldObjectManager
    inventory: InventoryManager
    login_data: Dict[str, Any]
    REGION_CLS = Type[BaseClientRegion]

    def __init__(self, id, secure_session_id, agent_id, circuit_code,
                 session_manager: Optional[BaseClientSessionManager], login_data=None):
        self.login_data = login_data or {}
        self.pending = True
        self.id: UUID = id
        self.secure_session_id: UUID = secure_session_id
        self.agent_id: UUID = agent_id
        self.circuit_code = circuit_code
        self.global_caps = {}
        self.session_manager = session_manager
        self.active_group: UUID = UUID.ZERO
        self.groups: Set[UUID] = set()
        self.regions = []
        self._main_region = None
        self.message_handler: MessageHandler[Message, str] = MessageHandler()
        super().__init__()

    @classmethod
    def from_login_data(cls, login_data, session_manager):
        sess = cls(
            id=UUID(login_data["session_id"]),
            secure_session_id=UUID(login_data["secure_session_id"]),
            agent_id=UUID(login_data["agent_id"]),
            circuit_code=int(login_data["circuit_code"]),
            session_manager=session_manager,
            login_data=login_data,
        )
        appearance_service = login_data.get("agent_appearance_service")
        map_image_service = login_data.get("map-server-url")
        if appearance_service:
            sess.global_caps["AppearanceService"] = appearance_service
        if map_image_service:
            sess.global_caps["MapImageService"] = map_image_service
        # Login data also has details about the initial sim
        sess.register_region(
            circuit_addr=(login_data["sim_ip"], login_data["sim_port"]),
            handle=(login_data["region_x"] << 32) | login_data["region_y"],
            seed_url=login_data["seed_capability"],
        )
        return sess

    def register_region(self, circuit_addr: Optional[ADDR_TUPLE] = None, seed_url: Optional[str] = None,
                        handle: Optional[int] = None) -> BaseClientRegion:
        if not any((circuit_addr, seed_url)):
            raise ValueError("One of circuit_addr and seed_url must be defined!")

        for region in self.regions:
            if region.circuit_addr == circuit_addr:
                if seed_url and region.cap_urls.get("Seed") != seed_url:
                    region.update_caps({"Seed": seed_url})
                if handle:
                    region.handle = handle
                return region
            if seed_url and region.cap_urls.get("Seed") == seed_url:
                return region

        if not circuit_addr:
            raise ValueError("Can't create region without circuit addr!")

        logging.info("Registering region for %r" % (circuit_addr,))
        region = self.REGION_CLS(circuit_addr, seed_url, self, handle=handle)
        self.regions.append(region)
        return region

    @property
    def main_region(self) -> Optional[BaseClientRegion]:
        if self._main_region and self._main_region() in self.regions:
            return self._main_region()
        return None

    @main_region.setter
    def main_region(self, val: BaseClientRegion):
        self._main_region = weakref.ref(val)

    def transaction_to_assetid(self, transaction_id: UUID):
        return UUID.combine(transaction_id, self.secure_session_id)

    def region_by_circuit_addr(self, circuit_addr) -> Optional[BaseClientRegion]:
        for region in self.regions:
            if region.circuit_addr == circuit_addr and region.circuit:
                return region
        return None

    def region_by_handle(self, handle: int) -> Optional[BaseClientRegion]:
        for region in self.regions:
            if region.handle == handle:
                return region
        return None

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.id)



import asyncio
import dataclasses
import unittest
from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import create_logged_task
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.templates import (
    AssetType,
    EstateAssetType,
    TransferSourceType,
    TransferRequestParamsSimEstate,
    TransferChannelType,
    TransferTargetType,
    TransferStatus,
)
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.base.transfer_manager import TransferManager, Transfer
from hippolyzer.lib.base.xfer_manager import XferManager
from hippolyzer.lib.base.test_utils import MockHandlingCircuit, MockConnectionHolder


class BaseTransferTests(unittest.IsolatedAsyncioTestCase):
    SMALL_PAYLOAD = b"foobar"
    LARGE_PAYLOAD = b"foobar" * 500

    def setUp(self) -> None:
        self.server_message_handler: MessageHandler[Message, str] = MessageHandler()
        self.client_message_handler: MessageHandler[Message, str] = MessageHandler()
        # The client side should send messages to the server side's message handler
        # and vice-versa
        self.client_circuit = MockHandlingCircuit(self.server_message_handler)
        self.server_circuit = MockHandlingCircuit(self.client_message_handler)
        self.server_connection = MockConnectionHolder(self.server_circuit, self.server_message_handler)
        self.client_connection = MockConnectionHolder(self.client_circuit, self.client_message_handler)


class XferManagerTests(BaseTransferTests):
    def setUp(self) -> None:
        super().setUp()
        self.secure_session_id = UUID.random()
        self.xfer_manager = XferManager(self.client_connection, self.secure_session_id)
        self.received_bytes: Optional[bytes] = None

    async def _handle_vfile_upload(self):
        msg = await self.server_message_handler.wait_for(('AssetUploadRequest',), timeout=0.01)
        asset_block = msg["AssetBlock"]
        transaction_id = asset_block["TransactionID"]
        asset_id = UUID.combine(transaction_id, self.secure_session_id)
        if asset_block["AssetData"]:
            self.received_bytes = asset_block["AssetData"]
        else:
            manager = XferManager(self.server_connection)
            xfer = await manager.request(vfile_id=asset_id, vfile_type=AssetType.BODYPART)
            self.received_bytes = xfer.reassemble_chunks()
        self.server_circuit.send(Message(
            "AssetUploadComplete",
            Block("AssetBlock", UUID=asset_id, Type=asset_block["Type"], Success=True),
            direction=Direction.IN,
        ))

    async def test_small_xfer_upload(self):
        _ = create_logged_task(self._handle_vfile_upload())
        await asyncio.wait_for(self.xfer_manager.upload_asset(
            AssetType.BODYPART, self.SMALL_PAYLOAD
        ), timeout=0.1)
        self.assertEqual(self.received_bytes, self.SMALL_PAYLOAD)

    async def test_large_xfer_upload(self):
        # Larger payloads take a different path
        _ = create_logged_task(self._handle_vfile_upload())
        await asyncio.wait_for(self.xfer_manager.upload_asset(
            AssetType.BODYPART, self.LARGE_PAYLOAD
        ), timeout=0.1)
        self.assertEqual(self.received_bytes, self.LARGE_PAYLOAD)


class TestTransferManager(BaseTransferTests):
    def setUp(self) -> None:
        super().setUp()
        self.transfer_manager = TransferManager(
            self.client_connection,
            UUID.random(),
            UUID.random(),
        )

    async def _handle_covenant_download(self):
        msg = await self.server_message_handler.wait_for(('TransferRequest',), timeout=0.01)
        self.assertEqual(TransferSourceType.SIM_ESTATE, msg["TransferInfo"]["SourceType"])
        tid = msg["TransferInfo"]["TransferID"]
        params: TransferRequestParamsSimEstate = msg["TransferInfo"][0].deserialize_var("Params")
        self.assertEqual(EstateAssetType.COVENANT, params.EstateAssetType)
        data = self.LARGE_PAYLOAD

        self.server_circuit.send(Message(
            'TransferInfo',
            Block(
                'TransferInfo',
                TransferID=tid,
                ChannelType=TransferChannelType.MISC,
                TargetType_=TransferTargetType.UNKNOWN,
                Status=TransferStatus.OK,
                Size=len(self.LARGE_PAYLOAD),
                Params_=dataclasses.asdict(params),
            ),
        ))
        packet_num = 0
        while True:
            chunk = data[:1000]
            data = data[1000:]
            self.server_circuit.send(Message(
                'TransferPacket',
                Block(
                    'TransferData',
                    TransferID=tid,
                    ChannelType=TransferChannelType.MISC,
                    Packet=packet_num,
                    Status=TransferStatus.OK if data else TransferStatus.DONE,
                    Data=chunk,
                )
            ))
            if not data:
                break
            packet_num += 1

    async def test_simple_transfer(self):
        _ = create_logged_task(self._handle_covenant_download())
        transfer: Transfer = await asyncio.wait_for(self.transfer_manager.request(
            source_type=TransferSourceType.SIM_ESTATE,
            params=TransferRequestParamsSimEstate(
                EstateAssetType=EstateAssetType.COVENANT,
            ),
        ), 0.1)
        self.assertEqual(len(self.LARGE_PAYLOAD), transfer.expected_size)
        self.assertEqual(self.LARGE_PAYLOAD, transfer.reassemble_chunks())



"""
Example anim mangler addon, to be used with local anim addon.

You can edit this live to apply various transforms to local anims,
as well as any uploaded anims. Any changes will be reflected in currently
playing local anims.

This example modifies any position keys of an animation's mHipRight joint.
"""
from hippolyzer.lib.base.llanim import Animation
from hippolyzer.lib.proxy.addons import AddonManager

import local_anim
AddonManager.hot_reload(local_anim, require_addons_loaded=True)


def offset_right_hip(anim: Animation):
    hip_joint = anim.joints.get("mHipRight")
    if hip_joint:
        for pos_frame in hip_joint.pos_keyframes:
            pos_frame.pos.Z *= 2.5
            pos_frame.pos.X *= 5.0
    return anim


class ExampleAnimManglerAddon(local_anim.BaseAnimManglerAddon):
    ANIM_MANGLERS = [
        offset_right_hip,
    ]


addons = [ExampleAnimManglerAddon()]



from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


def handle_lludp_message(session: Session, region: ProxiedRegion, message: Message):
    # addon_ctx will persist across addon reloads, use for storing data that
    # needs to survive across calls to this function
    ctx = session.addon_ctx[__name__]
    if message.name == "ChatFromViewer":
        chat = message["ChatData"]["Message"]
        if chat == "COUNT":
            ctx["chat_counter"] = ctx.get("chat_counter", 0) + 1
            message["ChatData"]["Message"] = str(ctx["chat_counter"])



import unittest

import aiohttp
import aioresponses
from yarl import URL

from hippolyzer.lib.base.network.caps_client import CapsClient


class TestCapsClient(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.caps = {}
        self.caps_client = CapsClient(self.caps)

    async def test_bare_url_works(self):
        with aioresponses.aioresponses() as m:
            m.get("https://example.com/", body=b"foo")
            async with self.caps_client.get("https://example.com/") as resp:
                self.assertEqual(await resp.read(), b"foo")

    async def test_own_session_works(self):
        with aioresponses.aioresponses() as m:
            async with aiohttp.ClientSession() as sess:
                m.get("https://example.com/", body=b"foo")
                async with self.caps_client.get("https://example.com/", session=sess) as resp:
                    self.assertEqual(await resp.read(), b"foo")

    async def test_read_llsd(self):
        with aioresponses.aioresponses() as m:
            m.get("https://example.com/", body=b"<llsd><integer>2</integer></llsd>")
            async with self.caps_client.get("https://example.com/") as resp:
                self.assertEqual(await resp.read_llsd(), 2)

    async def test_caps(self):
        self.caps.update({"Foobar": "https://example.com/"})
        with aioresponses.aioresponses() as m:
            m.post("https://example.com/baz", body=b"ok")
            data = {"hi": "hello"}
            headers = {"Foo": "bar"}
            async with self.caps_client.post("Foobar", path="baz", llsd=data, headers=headers) as resp:
                self.assertEqual(await resp.read(), b"ok")

            # Our original dict should not have been touched
            self.assertEqual(headers, {"Foo": "bar"})

            req_key = ("POST", URL("https://example.com/baz"))
            req_body = m.requests[req_key][0].kwargs['data']
            self.assertEqual(req_body, b'<?xml version="1.0" ?><llsd><map><key>hi</key><string>hello'
                                       b'</string></map></llsd>')

            with self.assertRaises(KeyError):
                with self.caps_client.get("BadCap"):
                    assert False



"""
Detect receipt of a marketplace order for a demo, and auto-attach the most appropriate object
"""

import asyncio
import re
from typing import List, Tuple, Dict, Optional, Sequence

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.templates import InventoryType, Permissions, FolderType
from hippolyzer.lib.proxy.addon_utils import BaseAddon, show_message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


MARKETPLACE_TRANSACTION_ID = UUID('ffffffff-ffff-ffff-ffff-ffffffffffff')


class DemoAutoAttacher(BaseAddon):
    def handle_eq_event(self, session: Session, region: ProxiedRegion, event: dict):
        if event["message"] != "BulkUpdateInventory":
            return
        # Check that this update even possibly came from the marketplace
        if event["body"]["AgentData"][0]["TransactionID"] != MARKETPLACE_TRANSACTION_ID:
            return
        # Make sure that the transaction targeted our real received items folder
        folders = event["body"]["FolderData"]
        received_folder = folders[0]
        if received_folder["Name"] != "Received Items":
            return
        skel = session.login_data['inventory-skeleton']
        actual_received = [x for x in skel if x['type_default'] == FolderType.INBOX]
        assert actual_received
        if UUID(actual_received[0]['folder_id']) != received_folder["FolderID"]:
            show_message(f"Strange received folder ID spoofing? {folders!r}")
            return

        if not re.match(r".*\bdemo\b.*", folders[1]["Name"], flags=re.I):
            return
        # Alright, so we have a demo... thing from the marketplace. What now?
        items = event["body"]["ItemData"]
        object_items = [x for x in items if x["InvType"] == InventoryType.OBJECT]
        if not object_items:
            return
        self._schedule_task(self._attach_best_object(session, region, object_items))

    async def _attach_best_object(self, session: Session, region: ProxiedRegion, object_items: List[Dict]):
        own_body_type = await self._guess_own_body(session, region)
        show_message(f"Trying to find demo for {own_body_type}")
        guess_patterns = self.BODY_CLOTHING_PATTERNS.get(own_body_type)
        to_attach = []
        if own_body_type and guess_patterns:
            matching_items = self._get_matching_items(object_items, guess_patterns)
            if matching_items:
                # Only take the first one
                to_attach.append(matching_items[0])
        if not to_attach:
            # Don't know what body's being used or couldn't figure out what item
            # would work best with our body. Just attach the first object in the folder.
            to_attach.append(object_items[0])

        # Also attach whatever HUDs, maybe we need them.
        for hud in self._get_matching_items(object_items, ("hud",)):
            if hud not in to_attach:
                to_attach.append(hud)

        region.circuit.send(Message(
            'RezMultipleAttachmentsFromInv',
            Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
            Block('HeaderData', CompoundMsgID=UUID.random(), TotalObjects=len(to_attach), FirstDetachAll=0),
            *[Block(
                'ObjectData',
                ItemID=o["ItemID"],
                OwnerID=session.agent_id,
                # 128 = "add", uses whatever attachmentpt was defined on the object
                AttachmentPt=128,
                ItemFlags_=(),
                GroupMask_=(),
                EveryoneMask_=(),
                NextOwnerMask_=(Permissions.COPY | Permissions.MOVE),
                Name=o["Name"],
                Description=o["Description"],
            ) for o in to_attach]
        ))

    def _get_matching_items(self, items: List[dict], patterns: Sequence[str]):
        # Loop over patterns to search for our body type, in order of preference
        matched = []
        for guess_pattern in patterns:
            # Check each item for that pattern
            for item in items:
                if re.match(rf".*\b{guess_pattern}\b.*", item["Name"], re.I):
                    matched.append(item)
        return matched

    # We scan the agent's attached objects to guess what kind of body they use
    BODY_PREFIXES = {
        "-Belleza- Jake ": "jake",
        "-Belleza- Freya ": "freya",
        "-Belleza- Isis ": "isis",
        "-Belleza- Venus ": "venus",
        "[Signature] Gianni Body": "gianni",
        "[Signature] Geralt Body": "geralt",
        "Maitreya Mesh Body - Lara": "maitreya",
        "Slink Physique Hourglass Petite": "hg_petite",
        "Slink Physique Mesh Body Hourglass": "hourglass",
        "Slink Physique Original Petite": "phys_petite",
        "Slink Physique Mesh Body Original": "physique",
        "[BODY] Legacy (f)": "legacy_f",
        "[BODY] Legacy (m)": "legacy_m",
        "[Signature] Alice Body": "sig_alice",
        "Slink Physique MALE Mesh Body": "slink_male",
        "AESTHETIC - [Mesh Body]": "aesthetic",
    }

    # Different bodies' clothes have different naming conventions according to different merchants.
    # These are common naming patterns we use to choose objects to attach, in order of preference.
    BODY_CLOTHING_PATTERNS: Dict[str, Tuple[str, ...]] = {
        "jake": ("jake", "belleza"),
        "freya": ("freya", "belleza"),
        "isis": ("isis", "belleza"),
        "venus": ("venus", "belleza"),
        "gianni": ("gianni", "signature", "sig"),
        "geralt": ("geralt", "signature", "sig"),
        "hg_petite": ("hourglass petite", "hg petite", "hourglass", "hg", "slink"),
        "hourglass": ("hourglass", "hg", "slink"),
        "phys_petite": ("physique petite", "phys petite", "physique", "phys", "slink"),
        "physique": ("physique", "phys", "slink"),
        "legacy_f": ("legacy",),
        "legacy_m": ("legacy",),
        "sig_alice": ("alice", "signature"),
        "slink_male": ("physique", "slink"),
        "aesthetic": ("aesthetic",),
    }

    async def _guess_own_body(self, session: Session, region: ProxiedRegion) -> Optional[str]:
        agent_obj = region.objects.lookup_fullid(session.agent_id)
        if not agent_obj:
            return None
        # We probably won't know the names for all of our attachments, request them.
        # Could be obviated by looking at the COF, not worth it for this.
        try:
            await asyncio.wait(region.objects.request_object_properties(agent_obj.Children), timeout=0.5)
        except asyncio.TimeoutError:
            # We expect that we just won't ever receive some property requests, that's fine
            pass

        for prefix, body_type in self.BODY_PREFIXES.items():
            for obj in agent_obj.Children:
                if not obj.Name:
                    continue
                if obj.Name.startswith(prefix):
                    return body_type
        return None


addons = [DemoAutoAttacher()]
